from __future__ import annotations

import asyncio
import concurrent.futures
import functools
import multiprocessing.queues
import queue
import threading
import weakref
from abc import ABC, abstractmethod
from collections.abc import Hashable
from datetime import datetime

from .concurrent.futures import ThreadPoolExecutor


class Task:
    """Objects of this class are created by ``BackgroundTask``."""

    _NOTSET_ = object()

    def __init__(
        self,
        task_id: Hashable,
        future: concurrent.futures.Future,
        callers: int,
        cancelled: threading.Event,
        info: queue.Queue | multiprocessing.queues.Queue,
        task_catalog: dict,
    ):
        self.task_id = task_id
        self.future = future
        self.callers = callers
        self._cancelled = cancelled
        self._info = info
        self._info_ = None
        self._task_catalog = task_catalog
        self._result_retrieved = 0

    def __del__(self):
        if not self.done() and not self.cancelled():
            self.cancel()

    def info(self):
        while True:
            # Get the last entry in the queue.
            try:
                self._info_ = self._info.get_nowait()
            except queue.Empty:
                break
        return self._info_

    def cancel(self, wait: bool = True) -> bool:
        if self.cancelled():
            return True

        self.callers -= 1
        if self.callers > 0:
            return False

        if self.future.done():
            return False

        self._cancelled.set()
        z = self.future.cancel()
        if not z:
            # Task has started execution, hence
            # `Future.cancel` can do nothing to stop it.
            # Now it relies on the user code in `BackgroundTask.run`
            # to check `_cancelled` and stop execution proactively.
            if wait:
                concurrent.futures.wait([self.future])
        del self._task_catalog[self.task_id]
        return True

    async def a_cancel(self, wait: bool = True) -> bool:
        z = self.cancel(False)
        if z and wait:
            while not self.future.done():
                await asyncio.sleep(0.0123)
        return z

    def done(self) -> bool:
        return self.future.done()

    def cancelled(self) -> bool:
        return self._cancelled.is_set()

    def result(self, timeout=None):
        if self.cancelled():
            raise concurrent.futures.CancelledError()

        err = None
        try:
            result = self.future.result(timeout=timeout)
        except concurrent.futures.TimeoutError:
            raise
        except Exception as e:
            err = e

        self._result_retrieved += 1
        if self._result_retrieved >= self.callers:
            # Remove from the manager if every stakeholder
            # has retrieved the result.
            try:
                del self._task_catalog[self.task_id]
            except KeyError:
                pass
        if err is not None:
            raise err
        return result

    async def a_result(self, timeout=None):
        try:
            return self.result(timeout=0)
        except concurrent.futures.TimeoutError:
            if timeout == 0:
                raise
            if timeout is None:
                timeout = 3600 * 24
            while not self.future.done():
                if timeout <= 0:
                    raise concurrent.futures.TimeoutError()
                await asyncio.sleep(0.0123)
                timeout -= 0.0123

            self._result_retrieved += 1
            if self._result_retrieved >= self.callers:
                # Remove from the manager if every stakeholder
                # has retrieved the result.
                try:
                    del self._task_catalog[self.task_id]
                except KeyError:
                    pass
            return self.future.result(0)

    def __await__(self):
        return self.a_result().__await__()

    def exception(self):
        """
        Returns ``None`` if task is not yet completed
        or has completed successfully.

        Raises ``CancelledError`` if the task was cancelled.

        If the task raised an exception, this method will
        raise the same exception.
        """
        if self.cancelled():
            raise concurrent.futures.CancelledError()

        try:
            return self.future.exception(timeout=0)
            # This returns `None` if no exception was raised.
        except concurrent.futures.TimeoutError:
            return None


class BackgroundTask(ABC):
    """
    ``BackgroundTask`` provides a mechanism to manage "tasks" run in a
    thread pool or process pool. Given a user-specified way to determine
    a task based on the task parameters, a repeat submission of an
    existing task (i.e. a task with the same ID) will simply get
    access to the existing task rather than making a new submission.

    Facilities are provided to check task status, and cancel an ongoing task.
    """

    def __init__(self, executor: concurrent.futures.Executor | None = None):
        """
        Parameters
        ----------
        executor
            If you provide your own *process* executor,
            it's recommended to use ``mpservice.concurrent.futures.ProcessPoolExecutor``.
        """
        self._own_executor = False
        if executor is None:
            executor = ThreadPoolExecutor()
            self._own_executor = True
        self._executor = executor
        self._tasks: dict[Hashable, Task] = {}

        if self._own_executor:
            weakref.finalize(self, executor.shutdown)

    @classmethod
    @abstractmethod
    def run(
        cls,
        *args,
        _cancelled: threading.Event,
        _info: queue.Queue | multiprocessing.queues.Queue,
        **kwargs,
    ):
        """
        This method contains the operations of the user task.
        It runs in ``self._executor``.

        In order to make sure these operations do not keep state
        in an instance of this class, it is a classmethod.

        Subclass implementation may want to make the parameter listing
        more specific.

        The parameters ``_cancelled`` and ``_info`` are provided by
        the background-task mechanism rather than the user task itself,
        but are to be used by the user task.

        If the task tends to take long, and wants to support cancellation,
        then it should check ``_cancelled.is_set()`` periodically (preferrably
        rather frequently), and stops if the flag is set.
        Note that ``concurrent.futures.Future.cancel``
        can only cancel the execution of a scheduled task if it has not started
        running yet. Once the task is ongoing, ``Future.cancel`` can't stop
        the execution. Therefore user code needs to be proactive, using
        ``_cancelled`` to detect the cancellation request and act accordingly.

        User would request cancellation using ``Task.cancel()``.

        The queue ``_info`` is used to pass info to the caller, e.g. periodic
        progress reports. Because there is no guarantee that this queue
        is being checked by the caller timely or at all,
        we can't let it have unlimited capacity and keep elements in it.
        In fact, it has length 1. Before pushing an element onto
        the queue, the user code should pop any existing elements.
        An element in this queue is typically a small dict.

        User would retrieve this info using ``Task.info()``.

        It is not mandatory that the user task code makes use of
        ``_cancelled`` and ``_info``.
        """
        raise NotImplementedError

    @classmethod
    def get_task_id(cls, *args, **kwargs) -> Hashable:
        """
        Determine an ID for the task based on the input parameters.
        Next time when a task is submitted, if its task ID
        turns out to be identical to an in-progress task,
        it will not be re-submitted; rather, it will just
        wait for the result of the in-progress task.

        There are some caveats in determination of this task ID.
        For example, if one of the input parameter is a large list,
        do we use the full content of the list to determine the ID?

        For another example, if the task takes the path of an input
        file, and the path is hard-coded and never changes, then
        this parameter value never changes; but does this mean
        the content of the file does not change?
        If we assume the file content could change anytime,
        we may consider each submission to constitute a unique task.
        Then, simply return a random task ID.

        The parameter list should be identical to that
        of ``run``, minus ``_cancelled`` and ``_status``.

        The default implementation returns a random value, meaning
        every new submission is considered a new, unique task.
        """
        return str(datetime.utcnow())

    def submit(self, *args, **kwargs) -> Task:
        task_id = self.get_task_id(*args, **kwargs)
        task = self._tasks.get(task_id)
        if task is not None:
            # The same task was submitted before.
            task.callers += 1
            if task.callers == 1:
                # This task was as "submit_and_forget".
                # Now that another user is submitting
                # the same task, and we don't know whether
                # this user also wants to "forget", we
                # need to remove the "forget" callback.
                try:
                    task.future._done_callbacks.remove(self._cb_remove_once_done)
                except ValueError:
                    pass
            return task

        if isinstance(self._executor, concurrent.futures.ThreadPoolExecutor):
            cancelled = threading.Event()
            info = queue.Queue(1)
        else:
            cancelled = self._executor._mp_context.Manager().Event()
            info = self._executor._mp_context.Manager().Queue(1)

        func = functools.partial(self.run, *args, **kwargs)
        fut = self._executor.submit(func, _cancelled=cancelled, _info=info)
        fut.task_id = task_id
        task = Task(
            task_id=task_id,
            future=fut,
            callers=1,
            cancelled=cancelled,
            info=info,
            task_catalog=self._tasks,
        )
        self._tasks[task_id] = task
        return task

    def _cb_remove_once_done(self, fut) -> None:
        del self._tasks[fut.task_id]

    def submit_and_forget(self, *args, **kwargs) -> None:
        task = self.submit(*args, **kwargs)
        if task.callers == 1:
            task.future.add_done_callback(self._cb_remove_once_done)
        task.callers -= 1

    def __contains__(self, task_id: str) -> bool:
        return task_id in self._tasks

    def __getitem__(self, task_id: str) -> Task:
        return self._tasks[task_id]

    def get(self, task_id: str, default=None) -> Task | None:
        return self._tasks.get(task_id, default)

    def __iter__(self):
        return iter(self._tasks)

    def keys(self):
        return self._tasks.keys()

    def values(self):
        return self._tasks.values()

    def items(self):
        return self._tasks.items()

    def __delitem__(self, task_id):
        del self._tasks[task_id]



import logging
import os
import signal
import socket
import sys
import warnings
from typing import Any, Callable, List, Optional, TypeVar
from weakref import WeakValueDictionary

import click
import uvicorn

from mpservice.multiprocessing import Event, Process

logger = logging.getLogger(__name__)


# See `uvicorn`.
class Server(uvicorn.Server):
    def run(
        self,
        *,
        sockets: list[socket.socket] | None = None,
        worker_context: Any = None,
        server_id: str,
        stop_event: Event = None,
    ) -> None:
        # When there are multiple worker processes, this method is the
        # `target` function that runs in a new process.
        config = self.config
        if not config.loaded:
            config.load()

        app = config.loaded_app
        while hasattr(app, 'app'):
            app = app.app
        assert not hasattr(app, 'worker_context')
        app.worker_context = worker_context
        # `app.worker_context` is designed to be used in the "lifespan"
        # of `app`. See Starlette documentation on "lifespan".

        global _servers
        if _servers is None:
            _servers = WeakValueDictionary()
        assert server_id not in _servers
        if stop_event is None:
            _servers[server_id] = self
        else:
            _servers[server_id] = stop_event
        self._stop_event = stop_event
        return super().run(sockets)

    async def on_tick(self, *args, **kwargs):
        return (await super().on_tick(*args, **kwargs)) or (
            self._stop_event is not None and self._stop_event.is_set()
        )


# See `uvicorn`.
class Multiprocess(uvicorn.supervisors.Multiprocess):
    def __init__(
        self, config, target, sockets, *, worker_contexts, server_id, stop_event
    ):
        super().__init__(config=config, target=target, sockets=sockets)
        self._worker_contexts = worker_contexts
        self._server_id = server_id
        self.should_exit = stop_event  # override the parent one

    def run(self) -> None:
        self.startup()
        self.should_exit.wait()
        self.shutdown()

    def startup(
        self,
    ) -> None:
        message = 'Started parent process [{}]'.format(str(self.pid))
        color_message = 'Started parent process [{}]'.format(
            click.style(str(self.pid), fg='cyan', bold=True)
        )
        logger.info(message, extra={'color_message': color_message})

        for sig in uvicorn.supervisors.multiprocess.HANDLED_SIGNALS:
            signal.signal(sig, self.signal_handler)

        for _idx in range(self.config.workers):
            process = get_subprocess(
                config=self.config,
                target=self.target,
                sockets=self.sockets,
                worker_context=self._worker_contexts[_idx],
                server_id=self._server_id,
                stop_event=self.should_exit,
                # Customization: pass in `worker_context`, `server_id`, and `stop_event`.
            )
            process.start()
            self.processes.append(process)

    def shutdown(self) -> None:
        for process in self.processes:
            # Customization: do not call `process.terminate()`
            # process.terminate()
            process.join()

        message = 'Stopping parent process [{}]'.format(str(self.pid))
        color_message = 'Stopping parent process [{}]'.format(
            click.style(str(self.pid), fg='cyan', bold=True)
        )
        logger.info(message, extra={'color_message': color_message})


# See `uvicorn`.
def get_subprocess(
    config: uvicorn.Config,
    target: Callable[..., None],
    sockets: List[socket.socket],
    worker_context: Any,
    server_id: str,
    stop_event: Event,
) -> Process:
    """
    Called in the parent process, to instantiate a new child process instance.
    The child is not yet started at this point.

    * config - The Uvicorn configuration instance.
    * target - A callable that accepts a list of sockets. In practice this will
               be the `Server.run()` method.
    * sockets - A list of sockets to pass to the server. Sockets are bound once
                by the parent process, and then passed to the child processes.
    """
    # We pass across the stdin fileno, and reopen it in the child process.
    # This is required for some debugging environments.

    # There are three customizations to this function:
    #   - use ``MP_SPAWN_CTX``
    #   - pass in ``worker_context``, ``server_id``, and ``stop_event``

    stdin_fileno: Optional[int]
    try:
        stdin_fileno = sys.stdin.fileno()
    except OSError:
        stdin_fileno = None

    kwargs = {
        'config': config,
        'target': target,
        'sockets': sockets,
        'stdin_fileno': stdin_fileno,
        'worker_context': worker_context,
        'server_id': server_id,
        'stop_event': stop_event,
    }

    return Process(target=subprocess_started, kwargs=kwargs)


# See `uvicorn`.
def subprocess_started(
    config: uvicorn.Config,
    target: Callable[..., None],
    sockets: List[socket.socket],
    stdin_fileno: Optional[int],
    worker_context: Any,
    server_id: str,
    stop_event: Event,
) -> None:
    """
    Called when the child process starts.

    * config - The Uvicorn configuration instance.
    * target - A callable that accepts a list of sockets. In practice this will
               be the `Server.run()` method.
    * sockets - A list of sockets to pass to the server. Sockets are bound once
                by the parent process, and then passed to the child processes.
    * stdin_fileno - The file number of sys.stdin, so that it can be reattached
                     to the child process.
    """
    # There is one customization to this function:
    #   - take and use ``worker_idx``

    # Re-open stdin.
    if stdin_fileno is not None:
        sys.stdin = os.fdopen(stdin_fileno)

    # Logging needs to be setup again for each child.
    config.configure_logging()

    # Now we can call into `Server.run(sockets=sockets)`
    target(
        sockets=sockets,
        worker_context=worker_context,
        server_id=server_id,
        stop_event=stop_event,
    )


async def stop_server(server_id: str = '0'):
    """
    This function is to be called in a web service endpoint to request termination
    of the service.
    """
    # If `start_server` was called with `workers=1`, then the `Server` object runs in the same process/thread
    # where `start_server` was called. This function is involked in the same process/thread, and we set the
    # simple `should_exit` flag to stop the server.
    # If `start_server` was called with `workers > 1`, then multiple `Server` objects run in their own processes.
    # This function is invoked in the process of one `Server`. If this function is called only once,
    # then only one `Server` (in the said process) will be stopped. To avoid asking client to call this multiple times,
    # We use a multiprocessing Event object shared between the server processes to get the signal across.

    s = _servers.get(server_id)
    if s is None:
        return
    if isinstance(s, Server):
        s.should_exit = True
    else:
        s.set()  # this is "propagated" to other worker processes as well as the ``Multiprocess`` object.


UNSET = object()

_servers: WeakValueDictionary[str, Server | Event] = None


ASGIApplication = TypeVar('ASGIApplication')


def start_server(
    app: ASGIApplication | str,
    *,
    host: str = '0.0.0.0',
    port: int = 8000,
    access_log: bool = False,
    backlog: int = 128,
    workers: int = 1,
    worker_contexts: list = None,
    log_config=UNSET,
    server_id: str = '0',
    **kwargs,
):
    """
    This function is intended for lauching a service that uses :mod:`mpservice.mpserver`.

    This function is adapted from ``uvicorn.main.run``.

    Parameters
    ----------
    app
        Usually `starlette.applications.Starlette <https://www.starlette.io/>`_ instance or the import string for such
        an instance, like ``'mypackage.mymodule:app'``. The module as named, `'mypackage.mymodule'`, must be able
        to be imported. If `app` is defined in a script (say `example.py`) that is in the current working directory,
        that is, if you type

            python example.py

        in the directory that contains `example.py`, then Python will be able to import that script.
        In that case, `app` can be written as `'example:app'`.

        If `workers > 1`, this must be a str, because a `Starlette` object cannot be pickled and sent
        to other processes.
    host
        The default ``'0.0.0.0'`` is suitable if the code runs within a Docker container.
    port
        Port.
    access_log
        Whether to let ``uvicorn`` emit access logs.
    backlog
        The default should be adequate. Don't make this large unless you know what you're doing.

        `uvicorn.Server.startup <https://github.com/encode/uvicorn/blob/master/uvicorn/server.py>`_
        passes this to AsyncIO ``loop.create_server`` (in
        `asyncio.base_events <https://github.com/python/cpython/blob/main/Lib/asyncio/base_events.py>`_,
        where default is 100), and in turn to
        `socket.socket.listen <https://docs.python.org/3/library/socket.html#socket.socket.listen>`_.

        ``uvicorn`` uses a large default value (2048). That might be misguided.

        See: https://bugs.python.org/issue38699, https://stackoverflow.com/a/2444491, https://stackoverflow.com/a/2444483,
    workers
        If ``1``, the server is launched in the current process/thread. If ``> 1``, this many processes
        are spawned, each running one copy of ``app`` independently.

        Note that worker processes are not dynamically created and killed according to need---this fixed
        number of processes are started and they remain active---this is in contrast to Gunicorn.
        This also makes the parameter ``worker_contexts`` meaningful.

        Since ``mpservice.mpserver`` handles sophisticated multiprocessing servers "natively",
        usually you should use ``workers=1``, that is, let ``uvicorn`` run a simple "in-process"
        worker (implemented by a ``mpservice.mpserver.AsyncServer`` object) and let ``AsyncServer``
        handle multiple worker processes.

        There *may* be cases where ``workers > 1`` has some speed advantage. One possible reason
        for such advantage might come from saving on pickling overhead. You need to bencharmk
        you particular use case to decide.
    worker_contexts
        If provided, this must be a list or tuple with as many as ``workers`` elements.
        The elements of ``worker_contexts`` will become the attribute ``worker_context`` of ``app``
        in the corresponding worker process (or the current process if ``workers=1``).

        This is mainly designed for passing config-style data to worker processes when ``workers > 1``.
        The ``app`` object is available to the "lifespan" of a ``Starlette`` object; user can obtain
        ``app.worker_context`` there and decide how to use it to initialize things, save it, as well
        as make it available to endpoint functions via the lifespan's "state". See tests for example usage.
    server_id
        If you run multiple services in the same program,
        you need to use `server_id` to distinguish them and provide the ID to both `start_server` and `stop_server`.

        However, that use case seems extrememly unlikely.
    **kwargs
        Passed to ``uvicorn.Config``.
    """
    if log_config is not UNSET:
        warnings.warn(
            '`log_config` is ignored since 0.14.1 and will be an error in 0.15.0',
            DeprecationWarning,
            stacklevel=2,
        )

    config = uvicorn.Config(
        app,
        host=host,
        port=port,
        access_log=access_log,
        workers=workers,
        backlog=backlog,
        log_config=None,
        **kwargs,
    )

    if not worker_contexts:
        worker_contexts = [None] * workers

    server = Server(config=config)
    if workers == 1:
        server.run(worker_context=worker_contexts[0], server_id=server_id)
    else:
        sock = config.bind_socket()
        mult = Multiprocess(
            config,
            target=server.run,
            sockets=[sock],
            worker_contexts=worker_contexts,
            server_id=server_id,
            stop_event=Event(),
        )
        mult.run()

    if config.uds and os.path.exists(config.uds):
        os.remove(config.uds)  # pragma: py-win32



from __future__ import annotations

import concurrent.futures
import errno
import logging
import logging.handlers
import multiprocessing
import multiprocessing.connection
import multiprocessing.context
import multiprocessing.managers
import multiprocessing.pool
import multiprocessing.queues
import multiprocessing.synchronize
import os
import time
import warnings
from multiprocessing import util
from typing import Generic, TypeVar

from .._common import TimeoutError
from ..threading import Thread
from .remote_exception import RemoteException

logger = logging.getLogger(__name__)


class SpawnProcess(multiprocessing.context.SpawnProcess):
    """
    A subclass of the standard ``multiprocessing.context.SpawnProcess``,
    this customization adds two things:

    1. Make result and exception available as attributes of the
       process object, hence letting you use a ``SpawnProcess`` object
       similarly to how you use the ``Future`` object returned by
       `concurrent.futures.ProcessPoolExecutor.submit <https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.Executor.submit>`_.
    2. Make logs in the worker process handled in the main process.

       Logging messages produced in worker processes are tricky.
       First, some settings should be concerned in the main process only,
       including log formatting, log-level control, log handler (destination), etc.
       Specifically, these should be settled in the "launching script", and definitely
       should not be concerned in worker processes.
       Second, the terminal printout of loggings in multiple processes tends to be
       intermingled and mis-ordered.

       This class uses a queue to transmit all logging messages that are produced
       in the worker process to the main process/thread, to be handled there.

    This class is aliased by ``mpservice.multiprocessing.Process``.
    It is preferred to import by the aliased name.

    Examples
    --------
    Let's use an example to show the logging behavior.
    First, use a spawn-context from the standard `multiprocessing`_:

    .. code-block:: python
        :linenos:

        # log.py
        import logging
        import multiprocessing as mp
        from mpservice.multiprocessing import SpawnProcess


        def worker():
            logging.getLogger('worker.error').error('worker error')
            logging.getLogger('worker.warn').warning('worker warning')
            logging.getLogger('worker.info').info('worker info')
            logging.getLogger('worker.debug').debug('worker debug')


        def main():
            logging.getLogger('main.error').error('main error')
            logging.getLogger('main.info').info('main info')
            p = mp.get_context('spawn').Process(target=worker)
            p.start()
            p.join()
            logging.getLogger('main.warn').warning('main warning')
            logging.getLogger('main.debug').debug('main debug')


        if __name__ == '__main__':
            logging.basicConfig(
                format='[%(asctime)s.%(msecs)02d; %(levelname)s; %(name)s; %(funcName)s, %(lineno)d] [%(processName)s]  %(message)s',
                level=logging.DEBUG,
            )
            main()

    Run it::

        $ python log.py
        [2022-12-20 17:29:54,386.386; ERROR; main.error; main, 15] [MainProcess]  main error
        [2022-12-20 17:29:54,386.386; INFO; main.info; main, 16] [MainProcess]  main info
        worker error
        worker warning
        [2022-12-20 17:29:54,422.422; WARNING; main.warn; main, 20] [MainProcess]  main warning
        [2022-12-20 17:29:54,423.423; DEBUG; main.debug; main, 21] [MainProcess]  main debug

    Clearly, the child process exhibits the default behavior---print the warning-and-above-level log messages to the console---unaware of the logging configuration set in the main process.
    **This is a show stopper.**

    On line 15, replace ``mp.get_context('spawn').Process`` by ``SpawnProcess``.
    Run it again::

        $ python log.py
        [2022-12-20 17:39:31,284.284; ERROR; main.error; main, 15] [MainProcess]  main error
        [2022-12-20 17:39:31,284.284; INFO; main.info; main, 16] [MainProcess]  main info
        [2022-12-20 17:39:31,321.321; ERROR; worker.error; worker, 8] [SpawnProcess-1]  worker error
        [2022-12-20 17:39:31,321.321; WARNING; worker.warn; worker, 9] [SpawnProcess-1]  worker warning
        [2022-12-20 17:39:31,321.321; INFO; worker.info; worker, 10] [SpawnProcess-1]  worker info
        [2022-12-20 17:39:31,322.322; DEBUG; worker.debug; worker, 11] [SpawnProcess-1]  worker debug
        [2022-12-20 17:39:31,327.327; WARNING; main.warn; main, 20] [MainProcess]  main warning
        [2022-12-20 17:39:31,327.327; DEBUG; main.debug; main, 21] [MainProcess]  main debug

    This time, logs in the child process respect the level and format configurations set in the main process
    (because they are sent to and handled in the main process).
    """

    def __init__(self, *args, kwargs=None, **moreargs):
        """
        Parameters
        ----------
        *args
            Positional arguments passed on to the standard ``Process``.
        kwargs
            Passed on to the standard ``Process``.
        **moreargs
            Additional keyword arguments passed on to the standard ``Process``.
        """
        if kwargs is None:
            kwargs = {}
        else:
            kwargs = dict(kwargs)

        assert '_result_and_error_' not in kwargs
        assert '_logger_queue_' not in kwargs
        reader, writer = multiprocessing.connection.Pipe(duplex=False)
        kwargs['_result_and_error_'] = writer
        logger_queue = MP_SPAWN_CTX.Queue()
        kwargs['_logger_queue_'] = logger_queue

        super().__init__(*args, kwargs=kwargs, **moreargs)

        assert not hasattr(self, '_result_and_error_')
        assert not hasattr(self, '_logger_queue_')
        assert not hasattr(self, '_logger_thread_')
        assert not hasattr(self, '_collector_thread_')
        assert not hasattr(self, '_finalize_')
        self._result_and_error_ = reader
        self._logger_queue_ = logger_queue
        self._logger_thread_ = None
        self._collector_thread_ = None
        self._finalize_ = None

    def start(self):
        super().start()

        # The following must be *after* ``super().start``, otherwise will get error
        # "cannot pickle '_thread.Lock' object" or "cannot pickle '_thread.RLock" because `self`
        # will be passed to the other process in ``super().start()``,
        # going through pickling.

        self._future_ = concurrent.futures.Future()

        self._logger_thread_ = Thread(
            target=self._run_logger_thread,
            args=(self._logger_queue_,),
            name='ProcessLoggerThread',
            daemon=True,
        )
        self._logger_thread_.start()

        self._collector_thread_ = Thread(
            target=self._run_collector_thread,
            name='ProcessCollectorThread',
        )
        self._collector_thread_.start()

        self._finalize_ = util.Finalize(
            self,
            type(self)._finalize_threads,
            (self._logger_thread_, self._logger_queue_, self._collector_thread_),
            exitpriority=5,
        )

    @staticmethod
    def _run_logger_thread(q: multiprocessing.queues.Queue):
        while True:
            record = q.get()
            if record is None:
                break
            logger = logging.getLogger(record.name)
            if record.levelno >= logger.getEffectiveLevel():
                logger.handle(record)

    def _run_collector_thread(self):
        result, error = None, None
        try:
            result = self._result_and_error_.recv()
            error = self._result_and_error_.recv()
        except EOFError as exc:
            # the process has been terminated by calling ``self.terminate()``
            while self.exitcode is None:
                time.sleep(0.001)
            exitcode = -self.exitcode
            if exitcode != 15:
                msg = os.strerror(exitcode)
                if exitcode == 9:
                    msg += ': possibly out of memory'
                error = OSError(exitcode, msg)
                error.__cause__ = exc
                print(f'Error in process "{self.name}": {error}')

        self._result_and_error_.close()
        self._result_and_error_ = None
        if error is not None:
            self._future_.set_exception(error)
        else:
            self._future_.set_result(result)
        self._logger_queue_.put(None)
        self._logger_thread_.join()

    @staticmethod
    def _finalize_threads(
        t_logger: Thread, q: multiprocessing.queues.Queue, t_collector: Thread
    ):
        q.put(None)
        t_logger.join()
        t_collector.join()

    @staticmethod
    def handle_exception(exc):
        # Subclass can customize this to log more info.
        print(f'{multiprocessing.current_process().name}: {repr(exc)}')

    def run(self):
        """
        Overrides the standard ``Process.run``.

        ``start`` arranges for this to be run in a child process.
        """
        result_and_error = self._kwargs.pop('_result_and_error_')

        # Upon completion, `result_and_error` will contain `result` and `exception`
        # in this order; both may be `None`.
        if self._target:
            logger_queue = self._kwargs.pop('_logger_queue_')

            if not logging.getLogger().hasHandlers():
                # Set up putting all log messages
                # ever produced in this process into ``logger_queue``,
                # The log messages will be consumed
                # in the main process by ``self._run_logger_thread``.
                #
                # During the execution of this process, logging should not be configured.
                # Logging config should happen in the main process/thread.
                root = logging.getLogger()
                root.setLevel(logging.DEBUG)
                qh = logging.handlers.QueueHandler(logger_queue)
                root.addHandler(qh)
                logging.captureWarnings(True)
            else:
                # If logger is configured in this process, then do not start log forwarding,
                # but this is usually not recommended.
                # This sually happends because logging is configured on the module level rather than
                # in the ``if __name__ == '__main__':`` block.
                logger_queue = None
                qh = None

            try:
                z = self._target(*self._args, **self._kwargs)
            except SystemExit:
                # TODO: what if `e.code` is not 0?
                result_and_error.send(None)
                result_and_error.send(None)
                raise  # should it raise or stay silent?
            except BaseException as e:
                self.handle_exception(e)
                # For some reason, this needs to go before the two lines below.
                # If this is after them, the customization in `mpservice.mpserver`
                # will not work---the extra log does not show.

                result_and_error.send(None)
                result_and_error.send(RemoteException(e))
                raise
            else:
                result_and_error.send(z)
                result_and_error.send(None)
            finally:
                result_and_error.close()
                if qh is not None:
                    logging.getLogger().removeHandler(qh)
                    logger_queue.close()
        else:
            result_and_error.send(None)
            result_and_error.send(None)
            result_and_error.close()

    def join(self, timeout=None):
        """
        Same behavior as the standard lib, except that if the process
        terminates with an exception, the exception is raised.
        """
        super().join(timeout=timeout)
        exitcode = self.exitcode
        if exitcode is None:
            # Not terminated. Timed out.
            return
        self._collector_thread_.join()
        if exitcode == 0:
            # Terminated w/o error.
            return
        if exitcode == 1:
            raise self._future_.exception()
        if exitcode >= 0:
            raise ValueError(f'expecting negative `exitcode` but got {exitcode}')
        exitcode = -exitcode
        if exitcode == errno.ENOTBLK:  # 15
            warnings.warn(
                f'process exitcode {exitcode}, {errno.errorcode[exitcode]}; likely due to a forced termination by calling `.terminate()`',
                stacklevel=2,
            )
            # For example, ``self.terminate()`` was called. That's a code smell.
            # ``signal.Signals.SIGTERM`` is 15.
            # ``signal.Signals.SIGKILL`` is 9.
            # ``signal.Signals.SIGINT`` is 2.
        else:
            if self._future_.exception():
                raise self._future_.exception()
            else:
                raise ChildProcessError(
                    f'child process failed with exitcode {exitcode}, {errno.errorcode[exitcode]}'
                )
        # For a little more info on the error codes, see
        #   https://www.gnu.org/software/libc/manual/html_node/Error-Codes.html

    def done(self) -> bool:
        """
        Return ``True`` if the process has terminated normally or with exception.
        Return ``False`` if the process is running or not yet started.
        """
        return self.exitcode is not None

    def result(self, timeout: float | int | None = None):
        """
        Behavior is similar to ``concurrent.futures.Future.result``.
        """
        super().join(timeout)
        if not self.done():
            raise TimeoutError
        return self._future_.result()

    def exception(self, timeout: float | int | None = None):
        """
        Behavior is similar to ``concurrent.futures.Future.exception``.
        """
        super().join(timeout)
        if not self.done():
            raise TimeoutError
        return self._future_.exception()


class SpawnContext(multiprocessing.context.SpawnContext):
    """
    The standard package ``multiprocessing`` has a
    `"context" <https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods>`_,
    which has to do with how a process is created and started.
    For example, a ``ForkContext`` (or ``SpawnContext``) will create processes by ``ForkProcess`` (or ``SpawnProcess``).
    To use a queue to communicate with the process, the queue needs to be created via ``ForkContext.Queue`` (or ``SpawnContext.Queue``),
    else it will not work.

    If you do

    ::

        import multiprocessing
        q = multiprocessing.Queue()

    this is equivalent to

    ::

        q = multiprocessing.get_context().Queue()

    `multiprocessing.get_context <https://docs.python.org/3/library/multiprocessing.html#multiprocessing.get_context>`_ takes the sole parameter ``method``,
    which **on Linux** defaults to ``'fork'``.

    However, it is advised to **not** use this default; rather, **always** use the "spawn" context.
    There are some references on this topic; for example, see `this article <https://pythonspeed.com/articles/python-multiprocessing/>`_
    and `this StackOverflow thread <https://stackoverflow.com/questions/64095876/multiprocessing-fork-vs-spawn>`_.

    So, multiprocessing code is better written this way::

        import multiprocessing
        ctx = multiprocessing.get_context('spawn')
        q = ctx.Queue(...)
        e = ctx.Event(...)
        p = ctx.Process(..., args=(q, e))
        ...

    where ``ctx`` is an instance of ``multiprocessing.context.SpawnContext``.

    ``mpservice.multiprocessing.SpawnContext`` inherits from the standard ``SpawnContext``
    and provides some enhancements. (The main enhancement is to use the custom :class:`mpservice.multiprocessing.SpawnProcess`
    for process creation.) The constant ``mpservice.multiprocessing.MP_SPAWN_CTX``
    points to an instance of this class. With these facilities, the code above can be replaced by this::

        from mpservice.multiprocessing import MP_SPAWN_CTX as ctx
        q = ctx.Queue(...)
        e = ctx.Event(...)
        p = ctx.Process(..., args=(q, e))
        ...

    However, there is an annoyance here: ``ctx.Queue``, ``ctx.Event``, and some others are not **classes**,
    but rather **factory methods**. As a result, they can not be used to annotate the type of the objects
    created by them. (Similarly, if you do ``from multiprocessing import Queue, Event``, you are getting
    factory methods, not classes. The actual classes, ``multiprocessing.queues.Queue`` and ``multiprocessing.synchronize.Event``,
    have a **required** parameter ``ctx``, hence are somewhat inconvenient to use.)

    The module ``mpservice.multiprocessing`` breaks from the standard ``multiprocessing`` API design to alleviate this problem.
    It provides custom classes :class:`Queue`, :class:`Event`, and some others, that have an **optional**
    rather than required parameter ``ctx``, which defaults to, you guessed it, ``mpservice.multiprocessing.MP_SPAWN_CTX``.
    With this, the code above is better yet written this way::

        from mpservice.multiprocessing import Queue, Event, Process
        q: Queue = Queue(...)
        e: Event = Event(...)
        p: Process = Process(..., args=(q, e))

    .. note:: Recommendations on the use of ``MP_SPAWN_CTX``: use the classes ``Process``, ``Manager``, ``Lock``,
        ``RLock``, ``Condition``, ``Semaphore``, ``BoundedSemaphore``, ``Event``, ``Barrier``,
        ``Queue``, ``JoinableQueue``, ``SimpleQueue``, ``Pool`` diretly to create objects and type-annote them;
        this is preferred over ``MP_SPAWN_CTX.Process``, ``MP_SPAWN_CTX.Manager``, etc, although they would work, too.
        A few other methods of ``SpawnContext`` are not exposed in ``mpservice.multiprocessing``;
        you may use them via the object ``MP_SPAWN_CTX``.
    """

    Process = SpawnProcess

    def Manager(self, *args, **kwargs):
        return SyncManager(*args, ctx=self.get_context(), **kwargs)

    def get_context(self, method=None):
        if method is None or method == 'spawn':
            return self
        return super().get_context(method)


# MP_SPAWN_CTX = multiprocessing.context.DefaultContext(SpawnContext())
# The version above would fail `tests/test_streamer.py::test_parmap`. I don't know why.
MP_SPAWN_CTX = SpawnContext()


class SyncManager(multiprocessing.managers.SyncManager):
    def __init__(
        self,
        *args,
        ctx=None,
        name: str | None = None,
        cpu: int | list[int] | None = None,
        **kwargs,
    ):
        super().__init__(*args, ctx=ctx or MP_SPAWN_CTX, **kwargs)
        self.start()
        if name:
            self._process.name = name
        if cpu is not None:
            if isinstance(cpu, int):
                cpu = [cpu]
            os.sched_setaffinity(self._process.pid, cpu)


class Lock(multiprocessing.synchronize.Lock):
    def __init__(self, *, ctx=None):
        super().__init__(ctx=ctx or MP_SPAWN_CTX)


class RLock(multiprocessing.synchronize.RLock):
    def __init__(self, *, ctx=None):
        super().__init__(ctx=ctx or MP_SPAWN_CTX)


class Condition(multiprocessing.synchronize.Condition):
    def __init__(self, lock=None, *, ctx=None):
        super().__init__(lock=lock, ctx=ctx or MP_SPAWN_CTX)


class Semaphore(multiprocessing.synchronize.Semaphore):
    def __init__(self, value=1, *, ctx=None):
        super().__init__(value=value, ctx=ctx or MP_SPAWN_CTX)


class BoundedSemaphore(multiprocessing.synchronize.BoundedSemaphore):
    def __init__(self, value=1, *, ctx=None):
        super().__init__(value=value, ctx=ctx or MP_SPAWN_CTX)


class Event(multiprocessing.synchronize.Event):
    def __init__(self, *, ctx=None):
        super().__init__(ctx=ctx or MP_SPAWN_CTX)


class Barrier(multiprocessing.synchronize.Barrier):
    def __init__(self, *args, ctx=None, **kwargs):
        super().__init__(*args, ctx=ctx or MP_SPAWN_CTX, **kwargs)


Elem = TypeVar('Elem')


class Queue(multiprocessing.queues.Queue, Generic[Elem]):
    def __init__(self, maxsize=0, *, ctx=None):
        super().__init__(maxsize, ctx=ctx or MP_SPAWN_CTX)


class JoinableQueue(multiprocessing.queues.JoinableQueue, Generic[Elem]):
    def __init__(self, maxsize=0, *, ctx=None):
        super().__init__(maxsize, ctx=ctx or MP_SPAWN_CTX)


class SimpleQueue(multiprocessing.queues.SimpleQueue, Generic[Elem]):
    def __init__(self, *, ctx=None):
        super().__init__(ctx=ctx or MP_SPAWN_CTX)


class Pool(multiprocessing.pool.Pool):
    @staticmethod
    def Process(*args, ctx=None, **kwargs):
        if ctx is None:
            ctx = MP_SPAWN_CTX
        return ctx.Process(*args, **kwargs)

    def __init__(self, *args, context=None, **kwargs):
        super().__init__(*args, context=context or MP_SPAWN_CTX, **kwargs)



"""
The standard `Managers <https://docs.python.org/3/library/multiprocessing.html#managers>`_
provide a mechanism to interactively
communicate with a "server" that runs in another process.

The keyword is "interactively"---just like a client to a HTTP server, the client process
can make a request to a "proxy" that is returned by the "manager" and wait for the response.
There may be multiple types of data to be requested, just like calling multiple "end-points" 
of a HTTP server.

The other inter-process data sharing facilities, e.g. a :class:`Queue`, work more in a passive fashion.
The client code checks the next data element in the queue, and obtaines whatever is there.

The interactivity is achieved by a "proxy": an object is created and *hosted* in a "server process";
the hosted object is represented by a "proxy", which is passed to and used in other processes
to communicate with the hosted object. The backbone of the communication is ``multiprocessing.connection.Connection``
with its ``send`` and ``recv`` methods.

The module ``mpservice.multiprocessing.server_process`` provides :class:`ServerProcess`,
with mainly two enhancements to the standard ``Manager``:
one concerns "shared memory",
and the other concerns returning a mix of hosted and non-hosted data from a proxy.
If you do not make use of these enhancements, you may as well use the 
:class:`Manager` that is imported
from ``mpservice.multiprocessing``.


Basic workflow
==============

1. First, register one or more classes with the :class:`ServerProcess` class::

    class Doubler:
        def __init__(self, ...):
            ...

        def double(self, x):
            return x * 2

    class Tripler:
        def __init__(self, ...):
            ...

        def triple(self, x):
            return x * 3

    ServerProcess.register('Doubler', Doubler)
    ServerProcess.register('Tripler', Tripler)

2. Then, start a "server process" object in a contextmanager::

    with ServerProcess() as server:
            ...

   By now, ``server`` has started a "server process". We interact with things
   running in the server process via methods that have been "registered" on ``server``.

3. Instances of the registered classes can be created within the server process::

        doubler = server.Doubler(...)
        tripler = server.Tripler(...)

   This causes corresponding class objects to be created
   in the server process; the returned objects are "proxies"
   for the real objects. These proxies can be passed to any other
   processes and used there.

   The arguments in the above calls are passed into the server process
   and used in the ``__init__`` methods of the corresponding classes.
   For this reason, the parameters to ``__init__`` of a registered class
   must all be pickle-able.

   Calling a registered class multiple times, like

   ::

        prox1 = server.Doubler(...)
        prox2 = server.Doubler(...)

   will create independent objects in the server process.

   Multiple ``ServerProcess`` objects will run their corresponding
   server processes independently.

4. Pass the proxy objects to any process and use them there.

   By default, public methods (minus "properties") defined by the registered classes
   can be invoked on a proxy and will return the expected
   result. For example,

   ::

        prox1.double(3)

   will return 6. Inputs and output of the method ``Doubler.double``
   must all be pickle-able.

   The computation of the proxy's methods happens in the server process.
   The proxy object only handles the communication and input/output transport.

   Between the server process and the proxy object in a particular client process or thread,
   a connection is established, which starts a new thread in the server process
   to handle all requests from that proxy object (in said client process or thread).
   These "requests" include calling all methods of the proxy, not just one particular method.

   For example,

   ::

        th = Thread(target=foo, args=(prox1,))
        th.start()
        ...

   Suppose in function ``foo`` the first parameter is called ``p`` (which is ``prox1`` passed in),
   then ``p`` will communicate with the ``Doubler`` object (which runs in the server process)
   via its new dedicated thread in the server process, separate from the connection thread for ``prox1``
   in the "current" thread.

   Consequently, calls on a particular method of the proxy from multiple processes or threads
   become multi-threaded concurrent calls in the server process.
   We can design a simple example to observe this effect::


        class Doubler:
            def do(self, x):
                time.sleep(0.5)
                return x + x

        ServerProcess.register('Doubler', Doubler)

        def main():
            with ServerProcess() as server:
                doubler = server.Doubler()

                ss = Stream(range(100)).parmap(doubler.do, executor='thread', num_workers=50)
                t0 = time.perf_counter()
                zz = list(ss)
                t1 = time.perf_counter()
                print(t1 - t0)
                assert zz == [x + x for x in range(100)]

        if __name__ == '__main__':
            main()

   If the calls to ``doubler.do`` were sequential, then 100 calls would take 50 seconds.
   With concurrent calls in 50 threads as above, it took 1.05 seconds in an experiment.

   It follows that, if the method mutates some shared state, you may need to use locks to guard things.

:class:`ServerProcess` inherits from the standard :class:`multiprocessing.managers.SyncManager`.
If you don't need the enhancements (except for using :class:`mpservice.multiprocessing.SpawnProcess`
in place of the standard :class:`multiprocessing.process.BaseProcess`), you can import ``Manager`` from
``mpservice.multiprocessing`` like this::

    from mpservice.multiprocessing import Manager

    with Manager() as manager:
        q = manager.Event()
        ...

Shared memory
=============

"Shared memory", introduced in Python 3.8, provides a way to use a block of memory across processes
with zero-copy. ``ServerProcess`` has a method :meth:`~mpservice.multiprocessing.server_process.ServerProcess.MemoryBlock`
that returns a :class:`~mpservice.multiprocessing.server_process.MemoryBlockProxy` object::

    with ServerProcess() as server:
        mem = server.MemoryBlock(1000)
        buffer = mem.buf  # memoryview
        # ... write data into `buffer`
        # pass `mem` to other processes and use its `.buf` again for the content.
        # Since it's a "shared" block of memory, any process can modify the data
        # via the memoryview.

Usually you will use the three methods 
:meth:`~mpservice.multiprocessing.server_process.MemoryBlockProxy.name`,
:meth:`~mpservice.multiprocessing.server_process.MemoryBlockProxy.size`,
and
:meth:`~mpservice.multiprocessing.server_process.MemoryBlockProxy.buf`,
of :class:`~mpservice.multiprocessing.server_process.MemoryBlockProxy`.
The property ``buf`` returns a ``memoryview`` of the shared memory.

The block of shared memory is released/destroyed once all references to the ``MemoryBlockProxy``
object have been garbage collected.
There are two special considerations to be aware of. First, in the "original" process
in the code block above, if ``mem`` is not deleted, it will be when ``server.__exit__``
executes.

Regarding the second consideration, suppose ``mem`` is placed in a Queue to be picked up
in another process; further suppose that the reference ``mem`` is gone after it has been placed
in the Queue. This happens, for example, ``mem` is a temporary reference in a loop.
In the standard behavior, the pickled data does not contain a reference to the ``MemoryBlockProxy``,
therefore there may very well be a moment when there is no reference to the shared memory outside
of the server process. However, the intention is clearly to use the shared memory once
the pickled ``MemoryBlockProxy`` is taken out of the Queue in another process.
For this reason, ``MemoryBlockProxy`` does some trick to count as a reference in pickled form.

Nested proxies
==============

The standard ``Manager`` allows "nested proxies" in a client process (not the "server process").
For example,

::

    with ServerProcess() as server:
        lst = server.list()
        dct = server.dict()
        dct['a'] = 3
        dct['b'] = 4
        lst.append(dct)

Now, the first element of ``lst`` is a dict proxy. You may pass ``lst`` to a nother process
and manipulate its first element; the changes are reflected in the dict that resides in the server process.

Return proxies from methods of hosted objects
=============================================

By default, when you call a method of a proxy, the returned value is plain old data passed from
the server process to the caller in the client process via pickling.
Once received, that value has no relation to the server process.

The standard ``Manager`` allows a method of a registered class to return a proxy
by instructions via the parameter ``method_to_typeid`` of :meth:`~mpservice.multiprocessing.server_process.ServerProcess.register`.
However, the set up can be inconvenient, and the capability is limited.

:class:`~mpservice.multiprocessing.server_process.ServerProcess` supports returning
a mix of proxy and non-proxy data elements in convenient and powerful ways.
Suppose we have a class ``Worker``, which we will run in a server process;
further, a method of ``Worker`` will return a dict that are plain data in some parts and
proxies in some others. We can do something like this::

    from mpservice.multiprocessing.server_process import hosted, MemoryBlock, ServerProcess

    class Worker:
        def make_data(self):
            return {
                'name': 'tom',
                'hobbies': ['soccor', 'swim'],
                'mem': hosted(MemoryBlock(100)),
                'jobs': ('IBM', 'APPLE', {'UK': hosted(['ARM', 'Shell'])}),
            }

        def get_memory(self, size):
            return hosted(MemoryBlock(size))

    ServerProcess.register('Worker', Worker, method_to_typeid={'make_data': 'hosted'})

Then we may use it like so::

    from mpservice.multiprocessing import Process

    def agent(data):
        assert data['name'] == 'tom'
        data['hobbies'].append('tennis')
        data['mem'].buf[3] = 8
        data['jobs][2]['UK'].append('BP')

    with ServerProcess() as server_process:
        worker = server_process.Worker()
        data = worker.make_data()
        p = Process(target=agent, args=(data, ))
        p.start()
        p.join()

        # data['hobbis'] is not hosted; it's modification in one process
        # will not be reflected in other processes.
        assert len(data['hobbies]) == 2

        assert data['mem'].buf[3] == 8
        assert data['jobs'][2]['UK'][2] == 'BP'

        mem = worker.get_memory(80)
        assert isinstance(mem, MomeryBlockProxy)
        assert mem.size == 80

Any value that is wrapped by ``hosted`` must be a registered class, e.g. the return of ``Worker.get_memory``.
When the entire return value is not wrapped by ``hosted``, the value must be a tuple, list, or dict.
Any element of a tuple or a list, or a value (not a key) in a dict, can be ``hosted``,
as long as that value is a ``register``\ed class. This goes recursively to any depth.
The search will stop at any value that is not a tuple, list, or dict; such a value will simply be pickled.
For example,

::

    data = {
        'a': [1, 2, [3, hosted(MemoryBlock)], ClassA()],
        'b': {'x': ['sfo', ('y', hosted([1, 2]))]},
    }
    return data

In this example, ``data['a'][2]`` is a list, hence we check each of its elements,
and find ``data['a'][2][1]`` to be a ``hosted`` object, but ``data['a'][2][2]``
is not a tuple, list, or dict, and not ``hosted``, hence it is returned as is.
Similarly, ``data['b']['x'][0]`` is a string and returned as is.
On the other hand, ``data['b']['x'][1]`` is a tuple, hence each of its elements is examined,
and it is found that it contains a plain string and a ``hosted`` list.
"""
from __future__ import annotations

import multiprocessing
import multiprocessing.connection
import multiprocessing.context
import multiprocessing.managers
import multiprocessing.pool
import multiprocessing.queues
import os
import sys
import threading
import traceback
import warnings
import weakref
from multiprocessing import util
from multiprocessing.managers import (
    BaseProxy,
    Token,
    convert_to_error,
    dispatch,
)
from traceback import format_exc

from ._context import MP_SPAWN_CTX

__all__ = [
    'ServerProcess',
    'MemoryBlock',
    'MemoryBlockProxy',
    'hosted',
]


# This is based on the 3.10 version.
def _callmethod(self, methodname, args=(), kwds={}):
    """
    Try to call a method of the referent and return a copy of the result
    """
    for tries in (0, 1):
        try:
            try:
                conn = self._tls.connection
            except AttributeError:
                util.debug(
                    'thread %r does not own a connection',
                    threading.current_thread().name,
                )
                self._connect()
                conn = self._tls.connection

            conn.send((self._id, methodname, args, kwds))
            break
        except BrokenPipeError as e:
            # In a few cases I saw ``BrokenPipeError: [Errno 32] Broken pipe``.
            # A workaround is described here:
            # https://stackoverflow.com/q/3649458/6178706

            if tries == 0:
                addr = self._token.address
                if addr in self._address_to_local:
                    del self._address_to_local[addr][0].connection
                    traceback.print_exc()
                    print(f'Retry once on "{e!r}"')
                    continue
            raise

    kind, result = conn.recv()

    if kind == '#RETURN':
        return result
    elif kind == '#PROXY':
        exposed, token = result
        if token.typeid == 'hosted':
            return HostedProxy(
                address=self._token.address,
                serializer=self._serializer,
                manager=self._manager,
                Client=self._Client,
                authkey=self._authkey,
                data=exposed,
            )

        proxytype = self._manager._registry[token.typeid][-1]
        token.address = self._token.address
        proxy = proxytype(
            token,
            self._serializer,
            manager=self._manager,
            authkey=self._authkey,
            exposed=exposed,
        )
        conn = self._Client(token.address, authkey=self._authkey)
        dispatch(conn, None, 'decref', (token.id,))
        return proxy
    raise convert_to_error(kind, result)


BaseProxy._callmethod = _callmethod


class PickleThroughProxy(BaseProxy):
    """
    With ``BaseProxy``, whenever a proxy object is garbage collected, ``decref`` is called
    on the hosted object in the server process. Once the refcount of the hosted object
    drops to 0, the object is discarded.

    If a proxy object is put in a queue and is removed from the code scope, then this proxy
    object is garbage collected once ``pickle.dumps`` is finished in the queue. The "data-in-pickle-form"
    of course does not contain a reference to the proxy object. This can lead to gabage collection
    of the object hosted in the server process.

    In some use cases, this may be undesirable. (See ``MemoryBlock`` and ``MemoryBlockProxy`` for an example.)
    This proxy classs is designed for this use case. It makes pickled data count as one reference
    to the server-process-hosted object.
    """

    def __reduce__(self):
        # The only sensible case of pickling this proxy object is for
        # transfering this object between processes in a queue.
        # (This proxy object is never pickled for storage.)
        #
        # Using ``MemoryBlock`` for example,
        # it is possible that at some time the only "reference"
        # to a certain shared memory block is in pickled form in
        # a queue. For example,
        #
        #   # main process
        #   with ServerProcess() as server:
        #        q = server.Queue()
        #        w = Process(target=..., args=(q, ))
        #        w.start()
        #
        #        for _ in range(1000):
        #           mem = server.MemoryBlock(100)
        #           # ... computations ...
        #           # ... write data into `mem` ...
        #           q.put(mem)
        #           ...
        #
        # During the loop, `mem` goes out of scope and only lives on in the queue.
        # A main design goal is to prevent the ``MemoryBlock`` object in the server process
        # from being garbage collected. This means we need to do some ref-counting hacks.

        # Inc ref here to represent the pickled object in the queue.
        # When unpickling, we do not inc ref again. In effect, we move the call
        # to ``incref`` earlier from ``pickle.loads`` into ``pickle.dumps``
        # for this object.

        conn = self._Client(self._token.address, authkey=self._authkey)
        dispatch(conn, None, 'incref', (self._id,))
        # NOTE:
        # calling ``self._incref()`` would not work because it adds another `_decref`;
        # I don't totally understand that part.

        func, args = super().__reduce__()
        args[-1]['incref'] = False  # do not inc ref again during unpickling.
        return RebuildPickleThroughProxy, (func, args)


def RebuildPickleThroughProxy(func, args):
    obj = func(*args)

    # We have called ``incref`` in ``PickleThroughProxy.__reduce__``,
    # i.e. prior to pickling,
    # hence we don't call ``incref`` when
    # reconstructing the proxy here during unpickling.
    # However, we still need to set up calling of ``decref``
    # when this reconstructed proxy is garbage collected.

    # The code below is part of ``BaseProxy._incref``.
    # ``BaseProxy.__init__`` does not do this.

    obj._idset.add(obj._id)
    state = obj._manager and obj._manager._state
    obj._close = util.Finalize(
        obj,
        BaseProxy._decref,
        args=(obj._token, obj._authkey, state, obj._tls, obj._idset, obj._Client),
        exitpriority=10,
    )

    return obj


def HostedProxy(*, address, serializer, manager, Client, authkey=None, data):
    def make_hosted(value: hosted):
        ident, exposed = value.value
        typeid = value.typeid

        tok = Token(typeid, address, ident)
        proxytype = manager._registry[typeid][-1]
        proxy = proxytype(
            tok, serializer, manager=manager, authkey=authkey, exposed=exposed
        )
        conn = Client(address, authkey=authkey)
        dispatch(conn, None, 'decref', (ident,))
        return proxy

    def make_one(value):
        if isinstance(value, hosted):
            return make_hosted(value)
        if isinstance(value, list):
            return make_list(value)
        if isinstance(value, dict):
            return make_dict(value)
        if isinstance(value, tuple):
            return make_tuple(value)
        return value

    def make_list(value: list):
        return [make_one(v) for v in value]

    def make_dict(value: dict):
        return {k: make_one(v) for k, v in value.items()}

    def make_tuple(value: tuple):
        return tuple(make_one(v) for v in value)

    return make_one(data)


class hosted:
    def __init__(self, value, typeid: str = None):
        self.value = value
        self.typeid = typeid or type(value).__name__


class _ProcessServer(multiprocessing.managers.Server):
    # This is the cpython code in versions 3.7-3.12.
    # My fix is labeled "FIX".
    def serve_client(self, conn):
        """
        Handle requests from the proxies in a particular process/thread
        """
        util.debug(
            'starting server thread to service %r', threading.current_thread().name
        )

        recv = conn.recv
        send = conn.send
        id_to_obj = self.id_to_obj

        Token = multiprocessing.managers.Token

        while not self.stop_event.is_set():
            try:
                methodname = obj = None
                request = recv()
                ident, methodname, args, kwds = request

                try:
                    obj, exposed, gettypeid = id_to_obj[ident]
                except KeyError as ke:
                    try:
                        obj, exposed, gettypeid = self.id_to_local_proxy_obj[ident]
                    except KeyError:
                        raise ke

                if methodname not in exposed:
                    raise AttributeError(
                        'method %r of %r object is not in exposed=%r'
                        % (methodname, type(obj), exposed)
                    )

                function = getattr(obj, methodname)
                # `function` carries a ref to ``obj``.

                try:
                    res = function(*args, **kwds)
                except Exception as e:
                    msg = ('#ERROR', e)
                else:
                    typeid = gettypeid and gettypeid.get(methodname, None)
                    if typeid:
                        rident, rexposed = self.create(conn, typeid, res)
                        token = Token(typeid, self.address, rident)
                        msg = ('#PROXY', (rexposed, token))
                    else:
                        msg = ('#RETURN', res)

                    # FIX
                    del res

                # FIX:
                # If no more request is coming, then `function` and `res` will stay around.
                # If `function` is a instance method of `obj`, then it carries a reference to `obj`.
                # Also, `res` is a refernce to the object that has been tracked in `self.id_to_obj`
                # and "returned" to the requester.
                # The extra reference to `obj` and `res` lingering here have no use, yet can cause
                # sutble bugs in applications that make use of their ref counts, such as ``MemoryBlock``.
                del function
                del obj

            except AttributeError:
                if methodname is None:
                    msg = ('#TRACEBACK', format_exc())
                else:
                    try:
                        fallback_func = self.fallback_mapping[methodname]
                        result = fallback_func(self, conn, ident, obj, *args, **kwds)
                        msg = ('#RETURN', result)

                        del obj  # HACK
                    except Exception:
                        msg = ('#TRACEBACK', format_exc())

            except EOFError:
                util.debug(
                    'got EOF -- exiting thread serving %r',
                    threading.current_thread().name,
                )
                sys.exit(0)

            except Exception:
                msg = ('#TRACEBACK', format_exc())

            try:
                try:
                    send(msg)
                except Exception:
                    send(('#UNSERIALIZABLE', format_exc()))
            except Exception as e:
                util.info(
                    'exception in thread serving %r', threading.current_thread().name
                )
                util.info(' ... message was %r', msg)
                util.info(' ... exception was %r', e)
                conn.close()
                sys.exit(1)

    def _create_hosted(self, c, data):
        # Construct the return value of a method that has been registered in
        # `method_to_typeid` as 'hosted'.

        def convert_hosted(value):
            value, typeid = value.value, value.typeid
            typeid_nocall = f'{typeid}-nocall'
            if typeid_nocall not in self.registry:
                self.registry[typeid_nocall] = (
                    None,
                    *self.registry[typeid][1:],
                )  # replace `callable` by `None`

            ident, exposed = self.create(c, typeid_nocall, value)
            # By now `value` has been referenced in ``self._id_to_obj``.
            return hosted((ident, exposed), typeid)

        def convert_one(value):
            if isinstance(value, hosted):
                return convert_hosted(value)
            if isinstance(value, list):
                return convert_list(value)
            if isinstance(value, dict):
                return convert_dict(value)
            if isinstance(value, tuple):
                return convert_tuple(value)
            return value  # any other type is returned as is, not hosted

        def convert_list(value):
            return [convert_one(v) for v in value]

        def convert_dict(value):
            return {k: convert_one(v) for k, v in value.items()}

        def convert_tuple(value):
            return tuple(convert_one(v) for v in value)

        return 0, convert_one(data)
        # The first value, `0`, is in the spot of ``ident`` but is just a placeholder.
        # The second value is the actual data and takes the spot of ``exposed``.
        # Refer to ``Server.serve_client``.
        # This data is received by ``BaseProxy._call_method``.

    def create(self, c, typeid, /, *args, **kwds):
        if typeid == 'hosted':
            assert not kwds
            assert len(args) == 1
            return self._create_hosted(c, args[0])
        return super().create(c, typeid, *args, **kwds)


class ServerProcess(multiprocessing.managers.SyncManager):
    # In each new thread or process, a proxy object will create a new
    # connection to the server process (see``multiprocessing.managers.Server.accepter``,
    # ...,
    # ``Server.accept_connection``,
    # and
    # ``BaseProxy._connect``;
    # all in `Lib/multiprocessing/managers.py <https://github.com/python/cpython/blob/main/Lib/multiprocessing/managers.py>`_);
    # the server process then creates a new thread
    # to handle requests from this connection (see ``Server.serve_client``
    # also in `Lib/multiprocessing/managers.py <https://github.com/python/cpython/blob/main/Lib/multiprocessing/managers.py>`_).

    # Overhead:
    #
    # I made a naive example, where the registered class just returns ``x + x``.
    # I was able to finish 13K calls per second.

    # Restrictions:
    # ``ServerProcess`` instances can't be pickled, hence can't be passed to other processes.
    # As a result, if you pass a proxy object to another process and call its methods in the other process,
    # the methods can't be ones that return proxy objects, because proxy objects would need a reference to
    # the "manager", while a proxy object does not carry its "manager" attribute during pickling.

    _Server = _ProcessServer

    @classmethod
    def register(
        cls,
        typeid: str,
        callable=None,
        **kwargs,
    ):
        """
        Typically, ``callable`` is a class object (not class instance) and ``typeid`` is the calss's name.

        Suppose ``Worker`` is a class, then registering ``Worker`` will add a method
        named "Worker" to the class ``ServerProcess``. Later on a running ServerProcess object
        ``server``, calling::

            server.Worker(*args, **kwargs)

        will run the callable ``Worker`` inside the server process, taking ``args`` and ``kwargs``
        as it normally would (since ``Worker`` is a class, treating it as a callable amounts to
        calling its ``__init__``). The object resulted from that call will stay in the server process.
        (In this, the call ``Worker(...)`` will result in an *instance* of the ``Worker`` class.)
        The call ``server.Worker(...)`` returns a "proxy" to that object; the proxy is going to be used
        from other processes or threads to communicate with the real object residing inside the server process.

        .. versionchanged:: 0.13.1
            Now the signature is consistent with that of :meth:`multiprocessing.managers.BaseManager.register`.

        .. note:: This method must be called before a :class:`ServerProcess` object is "started".
        """
        if not isinstance(typeid, str):
            assert callable is None
            assert not kwargs
            callable = typeid
            typeid = callable.__name__
            warnings.warn(
                'the signature of ``register`` has changed; now the first argument should be ``typeid``, which is a str',
                DeprecationWarning,
                stacklevel=2,
            )

        if typeid in cls._registry:
            warnings.warn(
                '"%s" is already registered; the existing registry is being overwritten.'
                % typeid
            )
        super().register(typeid=typeid, callable=callable, **kwargs)

    def __init__(
        self,
        *,
        cpu: int | list[int] | None = None,
        name: str | None = None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        name
            Name of the server process. If ``None``, a default name will be created.
        cpu
            Specify CPU pinning for the server process.
        """
        super().__init__(ctx=MP_SPAWN_CTX, **kwargs)
        self._name = name
        if isinstance(cpu, int):
            cpu = [cpu]
        self._cpu = cpu

    def __enter__(self):
        super().__enter__()
        if self._name:
            self._process.name = self._name

        if self._cpu is not None:
            os.sched_setaffinity(self._process.pid, self._cpu)

        self._memoryblock_proxies = weakref.WeakValueDictionary()
        return self

    def __exit__(self, *args):
        for prox in self._memoryblock_proxies.values():  # valuerefs():
            prox._close()
        # This takes care of dangling references to ``MemoryBlockProxy`` objects
        # in such situations (via ``ServerProcess``):
        #
        #   with ServerProcess() as server:
        #       mem = server.MemoryBlock(30)
        #       ...
        #       # the name ``mem`` is not "deleted" when exiting the context
        #
        # In this case, if the treatment above is not in place, we'll get such warning:
        #
        #   /usr/lib/python3.10/multiprocessing/resource_tracker.py:224: UserWarning: resource_tracker: There appear to be 1 leaked shared_memory objects to clean up at shutdown
        #
        # See ``MemoryBlockProxy.__init__()`` for how the object is registered in
        # ``self._memoryblock_proxies``.

        super().__exit__(*args)


ServerProcess.register(
    'hosted', callable=None, proxytype=HostedProxy, create_method=False
)


try:
    from multiprocessing.shared_memory import SharedMemory
except ImportError:
    pass
else:

    class MemoryBlock:
        """
        This class is used within the "server process" of a ``ServerProcess`` to
        create and track shared memory blocks.
        """

        _blocks_ = set()

        def __init__(self, size: int):
            assert size > 0
            self._mem = SharedMemory(create=True, size=size)
            self.__class__._blocks_.add(self._mem.name)

        def _name(self):
            # This is for use by ``MemoryBlockProxy``.
            return self._mem.name

        @property
        def name(self):
            return self._mem.name

        @property
        def size(self):
            # This could be slightly larger than the ``size`` passed in to :meth:`__init__`.
            return self._mem.size

        @property
        def buf(self):
            return self._mem.buf

        def _list_memory_blocks(self):
            """
            Return a list of the names of the shared memory blocks
            created and tracked by this class.
            This list includes the memory block created by the current
            ``MemoryBlock`` instance; in other words, the list
            includes ``self.name``.

            This is for use by ``MemoryBlockProxy``.
            This is mainly for debugging purposes.
            """

            return list(self._blocks_)

        def __del__(self):
            """
            The garbage collection of this object happens when its refcount
            reaches 0. Unless this object is "attached" to anything within
            the server process, it is garbage collected once all references to its
            corresponding proxy object (outside of the server process) have been
            removed.

            The current object creates a shared memory block and references it
            by a ``SharedMemory`` object. In addition, each proxy object creates
            a ``SharedMemory`` object pointing to the same memory block for
            reading/writing.
            According to the class ``multiprocessing.shared_memory.SharedMemory``,
            a ``SharedMemory`` object calls ``.close`` upon its garbage collection,
            hence when all the proxy objects goes away, all the ``SharedMemory`` objects
            they have created have called ``.close``.
            At that point, ``self._mem``, residing in the server process, is the only
            reference to the shared memory block, and it is safe to "destroy" the memory block.

            Therefore, this ``__del__`` destroys that memory block.
            """
            name = self._mem.name
            self._mem.close()
            self._mem.unlink()
            self.__class__._blocks_.remove(name)

        def __repr__(self):
            return f'<{self.__class__.__name__} {self.name()}, size {self.size()}>'

    class MemoryBlockProxy(PickleThroughProxy):
        _exposed_ = ('_list_memory_blocks', '_name')

        def __init__(self, *args, incref=True, manager=None, **kwargs):
            super().__init__(*args, incref=incref, manager=manager, **kwargs)
            self._name = None
            self._size = None
            self._mem = None

            if incref:
                # Since ``incref`` is ``True``,
                # ``self`` is being constructed out of the return of either
                # ``ServerProcess.MemoryBlock`` or a method of some proxy (
                # and the method's return value contains a ``MemoryBlock``).
                # ``self`` is not being constructed due to unpickling.
                #
                # We only place such objects under this tracking, and don't place
                # unpickled objects (in other processes) under this tracking, because
                # those objects in other processes should explicitly reach end of life.
                if self._manager is not None:
                    self._manager._memoryblock_proxies[self._id] = self

        def __reduce__(self):
            func, args = super().__reduce__()
            return _rebuild_memory_block_proxy, (
                func,
                args,
                self._name,
                self._size,
                self._mem,
            )

        @property
        def name(self):
            """
            Return the name of the ``SharedMemory`` object.
            """
            if self._name is None:
                self._name = self._callmethod('_name')
            return self._name

        @property
        def size(self) -> int:
            """
            Return size of the memory block in bytes.
            """
            if self._mem is None:
                self._mem = SharedMemory(name=self.name, create=False)
            return self._mem.size

        @property
        def buf(self) -> memoryview:
            """
            Return a ``memoryview`` into the context of the memory block.
            """
            if self._mem is None:
                self._mem = SharedMemory(name=self.name, create=False)
            return self._mem.buf

        def _list_memory_blocks(self) -> list[str]:
            """
            Return names of shared memory blocks being
            tracked by the ``ServerProcess`` that "ownes" the current
            proxy object.
            """
            return self._callmethod('_list_memory_blocks')

        def __str__(self):
            return f"<{self.__class__.__name__} '{self.name}' at {self._id}, size {self.size}>"

    def _rebuild_memory_block_proxy(func, args, name, size, mem):
        obj = func(*args)
        obj._name, obj._size, obj._mem = name, size, mem
        return obj

    ServerProcess.register('MemoryBlock', MemoryBlock, proxytype=MemoryBlockProxy)

    def list_memory_blocks(self):
        """
        List names of MemoryBlock objects being tracked in the server process.
        """
        m = self.MemoryBlock(1)
        blocks = m._list_memory_blocks()
        blocks.remove(m.name)
        return blocks

    ServerProcess.list_memory_blocks = list_memory_blocks



# Iterable vs iterator
#
# if we need to use
#
#   for v in X:
#       ...
#
# then `X.__iter__()` is called to get an "iterator".
# In this case, `X` is an "iterable", and it must implement `__iter__`.
#
# If we do not use it that way, but rather only directly call
#
#   next(X)
#
# then `X` must implement `__next__` (but does not need to implement `__iter__`).
# This `X` is an "iterator".
#
# Often we let `__iter__` return `self`, and implement `__next__` in the same class.
# This way the class is both an iterable and an iterator.

# Async generator returns an async iterator.

from __future__ import annotations

# Enable using a class in type annotations in the code
# that defines that class itself.
# https://stackoverflow.com/a/49872353
# Will no longer be needed in Python 3.10.
import asyncio
import concurrent.futures
import contextlib
import functools
import inspect
import itertools
import logging
import queue
import random
import threading
import traceback
from collections import deque
from collections.abc import AsyncIterable, AsyncIterator, Iterable, Iterator, Sequence
from inspect import iscoroutinefunction
from time import perf_counter
from types import SimpleNamespace
from typing import (
    Any,
    Awaitable,
    Callable,
    Generic,
    Literal,
    Optional,
    TypeVar,
)

import asyncstdlib.itertools
from typing_extensions import Self  # In 3.11, import this from `typing`

from ._queues import SingleLane
from .concurrent.futures import (
    ProcessPoolExecutor,
    ThreadPoolExecutor,
    get_shared_process_pool,
    get_shared_thread_pool,
)
from .multiprocessing import Event
from .multiprocessing.remote_exception import get_remote_traceback, is_remote_exception
from .threading import Thread

logger = logging.getLogger(__name__)

FINISHED = '8d906c4b-1161-40cc-b585-7cfb012bca26'
STOPPED = 'ceccca5e-9bb2-46c3-a5ad-29b3ba00ad3e'
NOTSET = object()


T = TypeVar('T')  # indicates input data element
TT = TypeVar('TT')  # indicates output after an op on `T`
Elem = TypeVar('Elem')


def isiterable(x):
    try:
        iter(x)
        return True
    except (TypeError, AttributeError):
        return False


def isasynciterable(x):
    try:
        aiter(x)
        return True
    except (TypeError, AttributeError):
        return False


class Stream(Generic[Elem]):
    """
    The class ``Stream`` is the "entry-point" for the "streamer" utilities.
    User constructs a ``Stream`` object
    by passing an `Iterable`_ or `AsyncIterable_` to it, then calls its methods to use it.
    Most of the methods return the object itself, facilitating calls
    in a "chained" fashion, like this::

        s = Stream(...).map(...).filter(...).batch(...).parmap(...).ubatch(...)

    However, these methods modify the object in-place, hence the above is equivalent
    to calling the methods one by one::

        s = Stream(...)
        s.map(...)
        s.filter(...)
        s.batch(...)
        s.parmap(...)
        s.unbatch(...)
    """

    def __init__(self, instream: Iterable | AsyncIterable, /):
        """
        Parameters
        ----------
        instream
            The input stream of elements, possibly unlimited.

            The sync-ness or async-ness of ``instream`` suggests to the ``Stream`` object
            whether the host environment is sync or async.
            In an async context, you may pass in a sync ``instream`` (hence creating
            a sync Stream), then turn it async like this::

                stream = Stream(range(1000)).to_async()
        """
        self.streamlets: list[Iterable | AsyncIterable] = [instream]

    def to_sync(self):
        """
        Make the stream "sync" iterable only.
        """
        if not isiterable(self):
            self.streamlets.append(SyncIter(self.streamlets[-1]))
        return self

    def to_async(self):
        """
        Make the stream "async" iterable only.
        """
        if not isasynciterable(self):
            self.streamlets.append(AsyncIter(self.streamlets[-1]))
        return self

    def __iter__(self) -> Iterator[Elem]:
        # Usually between ``__iter__`` and ``__aiter__`` only
        # one is available, dependending on the type of
        # ``self.streamlets[-1]``.
        return self.streamlets[-1].__iter__()

    def __aiter__(self) -> AsyncIterator[Elem]:
        return self.streamlets[-1].__aiter__()

    def drain(self) -> int:
        """Drain off the stream.

        If ``self`` is sync, return the number of elements processed.
        If ``self`` is async, return an awaitable that will do what the sync version does.

        This method is for the side effect: the entire stream has been processed
        by all the operations and results have been taken care of, for example,
        the final operation may have saved results in a database.

        If you need more info about the processing, such as inspecting exceptions
        as they happen (if ``return_expections`` to :meth:`parmap` is ``True``),
        then don't use this method. Instead, iterate the streamer yourself
        and do whatever you need to do.
        """
        if isiterable(self):
            n = 0
            for _ in self:
                n += 1
            return n

        async def main():
            n = 0
            async for _ in self:
                n += 1
            return n

        return main()

    def collect(self) -> list[Elem]:
        """If ``self`` is sync, return all the elements in a list.
        If ``self`` is async, return an awaitable that will do what the sync version does.

        .. warning:: Do not call this method on "big data".
        """
        if isiterable(self):
            return list(self)

        async def main():
            return [x async for x in self]

        return main()

    def _choose_by_mode(self, sync_choice, async_choice):
        if isiterable(self):
            return sync_choice
        return async_choice

    def map(self, func: Callable[[T], Any], /, **kwargs) -> Self:
        """Perform a simple transformation on each data element.

        This operation happens "inline"--there is no other threads or processes
        used. For this reason, the method is for "light-weight" transforms.

        This is a 1-to-1 transform from the input stream to the output stream.
        This method can neither add nor skip elements in the stream.

        If the logic needs to keep some state or history info, then define a class and implement
        its ``__call__`` method.
        For example, to print out every hundredth value for information::

            class Peek:
                def __init__(self):
                    self._count = 0

                def __call__(self, x):
                    self._count += 1
                    if self._count == 100:
                        print(x)
                        self._count = 0
                    return x

            obj.map(Peek())

        (This functionality is already provided by :meth:`peek`.)

        Parameters
        ----------
        func
            A function that takes a data element and returns a new value; the new values
            (which do not have to differ from the original) form the new stream going
            forward.

        *kwargs
            Additional keyword arguments to ``func``, after the first argument, which
            is the data element.
        """
        cls = self._choose_by_mode(Mapper, AsyncMapper)
        self.streamlets.append(cls(self.streamlets[-1], func, **kwargs))
        return self

    def filter(self, func: Callable[[T], bool], /, **kwargs) -> Self:
        """Select data elements to keep in the stream according to the predicate ``func``.

        This method can be used to either "keep" or "drop" elements according to various conditions.
        If the logic needs to keep some state or history info, then define a class and implement
        its ``__call__`` method.
        For example, to "drop the first 100 elements"::

            class Drop:
                def __init__(self, n):
                    self._n = n
                    self._count = 0

                def __call__(self, x):
                    self._count += 1
                    if self._count <= self._n:
                        return False
                    return True

            obj.filter(Drop(100))

        Parameters
        ----------
        func
            A function that takes a data element and returns ``True`` or ``False`` to indicate
            the element should be kept in or dropped from the stream.

            This function should not make changes to the input data element.
        *kwargs
            Additional keyword arguments to ``func``, after the first argument, which
            is the data element.
        """
        cls = self._choose_by_mode(Filter, AsyncFilter)
        self.streamlets.append(cls(self.streamlets[-1], func, **kwargs))
        return self

    def filter_exceptions(
        self,
        drop_exc_types: Optional[
            type[BaseException] | tuple[type[BaseException], ...]
        ] = None,
        keep_exc_types: Optional[
            type[BaseException] | tuple[type[BaseException], ...]
        ] = None,
    ) -> Self:
        """
        If a call to :meth:`parmap` upstream has specified ``return_exceptions=True``,
        then its output stream may contain ``Exception`` objects.
        Other methods such as :meth:`map` may also deliberately capture and return
        Exception objects.

        ``filter_exceptions`` determines which ``Exception`` objects in the stream
        should be dropped, kept, or raised.

        While :meth:`parmap` and other operators can choose to continue processing the stream
        by returning rather than raising exceptions,
        a subsequent ``filter_exceptions`` drops known types of exception objects
        so that the next operator does not receive ``Exception`` objects as inputs.

        The default behavior (both ``drop_exc_types`` and ``keep_exc_types`` are ``None``)
        is to raise any Exception object that is encountered.

        A useful pattern is to specify one or a few known exception types to drop,
        and crash on any other unexpected exception.

        Parameters
        ----------
        drop_exc_types
            These types of exceptions are dropped from the stream.
            These should be one or a few carefully identified exception types that
            you know can be safely ignored.

            If ``None`` (the default) or ``()`` or ``[]``, no exception object is dropped.

            To drop all exceptions, use ``Exception`` or even ``BaseException``.
        keep_exc_types
            These types of exceptions are kept in the stream.

            If ``None`` (the default), no exception object is kept.

            To keep all exceptions, use ``Exception`` or even ``BaseException``.

            An exception object that is neither kept nor dropped will be raised.

            .. note:: The members in ``keep_exc_types`` and ``drop_exc_types`` should be distinct.
                If there is any common member, then it is kept because the ``keep_exc_types``
                condition is checked first.
        """

        def foo(x):
            if isinstance(x, BaseException):
                if keep_exc_types is not None and isinstance(x, keep_exc_types):
                    return True
                if drop_exc_types is not None and isinstance(x, drop_exc_types):
                    return False
                raise x
            return True

        return self.filter(foo)

    def peek(
        self,
        *,
        print_func: Optional[Callable[[str], None]] = None,
        interval: int | float = 1,
        exc_types: Optional[Sequence[type[BaseException]]] = BaseException,
        with_exc_tb: bool = True,
        prefix: str = '',
        suffix: str = '',
    ) -> Self:
        """Take a peek at the data element *before* it continues in the stream.

        This is implemented by :meth:`map`, where the mapper function prints
        some info under certain conditions before returning the input value unchanged.
        If this does not do what you need, just create your own
        function to pass to :meth:`map`.

        Parameters
        ----------
        print_func
            A function that will be used to print messages.
            This should take a str and return nothing.

            The default is the built-in ``print``. It's often useful
            to pass in logging function such as ``logger.info``.
        interval
            Print out the data element at this interval. The default is 1000,
            that is, print every 1000th elment.

            If it is a float, then it must be between 0 and 1 open-open.
            This will be take as the (target) fraction of elements that are printed.

            To print out every element, pass in ``1``.

            To turn off the printouts by this condition, pass in ``None``.
        exc_types
            If the element is an exception object of these types, print it out.
            This is regardless of the ``interval`` value.

            If ``BaseException`` (the default), all exception objects are printed.

            To turn off the printouts by this condition, pass in ``None`` or ``()`` or ``[]``.
            In this case, ``interval`` is the only creterion for determining whether an element
            should be printed, the element being an ``Exception`` or not.
        with_exc_tb
            If ``True``, traceback, if available, will be printed when an ``Exception`` object is printed,
            the printing being determined by either ``interval`` or ``exc_types``.
        """
        if interval is not None:
            if isinstance(interval, float):
                assert 0 < interval < 1
            else:
                assert isinstance(interval, int)
                assert interval >= 1
        if exc_types is None:
            exc_types = ()
        elif type(exc_types) is type:  # a class
            exc_types = (exc_types,)
        if prefix:
            if not prefix.endswith(' ') and not prefix.endswith('\n'):
                prefix = prefix + ' '
        if suffix:
            if not suffix.startswith(' ') and not suffix.startswith('\n'):
                suffix = ' ' + suffix

        class Peeker:
            def __init__(self):
                self._idx = 0
                self._print_func = print if print_func is None else print_func
                self._interval = interval
                self._exc_types = exc_types
                self._with_trace = with_exc_tb
                self._prefix = prefix
                self._suffix = suffix

            def __call__(self, x):
                self._idx += 1
                should_print = False
                if self._interval is not None:
                    if self._interval >= 1:
                        if self._idx % self._interval == 0:
                            should_print = True
                    else:
                        should_print = random.random() < self._interval
                if (
                    not should_print
                    and self._exc_types
                    and isinstance(x, BaseException)
                    and isinstance(x, self._exc_types)
                ):
                    should_print = True
                if not should_print:
                    return x
                if not isinstance(x, BaseException):
                    self._print_func(f'{self._prefix}#{self._idx}:')
                    self._print_func(f'{x}{self._suffix}')
                    return x
                trace = ''
                if self._with_trace:
                    if is_remote_exception(x):
                        trace = get_remote_traceback(x)
                    else:
                        try:
                            trace = ''.join(traceback.format_tb(x.__traceback__))
                        except AttributeError:
                            pass
                self._print_func(f'{self._prefix}#{self._idx}:')
                if trace:
                    self._print_func(f'{x}')
                    self._print_func(f'{trace}{self._suffix}')
                else:
                    self._print_func(f'{x}{self._suffix}')
                return x

        return self.map(Peeker())

    def shuffle(self, buffer_size: int = 1000) -> Self:
        """
        Shuffle the elements using a buffer as intermediate storage.

        If ``buffer_size`` is ``>=`` the number of the elements, then the shuffling is fully random.
        Otherwise, the shuffling is somewhat "localized".
        """
        cls = self._choose_by_mode(Shuffler, AsyncShuffler)
        self.streamlets.append(cls(self.streamlets[-1], buffer_size=buffer_size))
        return self

    def head(self, n: int) -> Self:
        """
        Take the first ``n`` elements and ignore the rest.
        If the entire stream has less than ``n`` elements, just take all of them.

        This does not delegate to ``filter``, because ``filter``
        would need to walk through the entire stream,
        which is not needed for ``head``.
        """
        cls = self._choose_by_mode(Header, AsyncHeader)
        self.streamlets.append(cls(self.streamlets[-1], n))
        return self

    def tail(self, n: int) -> Self:
        """
        Take the last ``n`` elements and ignore all the previous ones.
        If the entire stream has less than ``n`` elements, just take all of them.

        .. note:: ``n`` data elements need to be kept in memory, hence ``n`` should
            not be "too large" for the typical size of the data elements.
        """
        cls = self._choose_by_mode(Tailor, AsyncTailor)
        self.streamlets.append(cls(self.streamlets[-1], n))
        return self

    def groupby(self, key: Callable[[T], Any], /, **kwargs) -> Self:
        """
        ``key`` takes a data element and outputs a value.
        **Consecutive** elements that have the same value of this output
        will be yielded as a generator.

        Following this operator, every element in the output stream is a tuple
        with the key and the values (as a generator).

        This function is similar to the standard
        `itertools.groupby <https://docs.python.org/3/library/itertools.html#itertools.groupby>`_.

        The output of ``key`` is usually a str or int or a tuple of a few strings or ints.

        ``**kwargs`` are additional keyword arguments to ``key``.

        Examples
        --------
        >>> data = ['atlas', 'apple', 'answer', 'bee', 'block', 'away', 'peter', 'question', 'plum', 'please']
        >>> print(Stream(data).groupby(lambda x: x[0]).map(lambda x: list(x[1])).collect())
        [['atlas', 'apple', 'answer'], ['bee', 'block'], ['away'], ['peter'], ['question'], ['plum', 'please']]
        """
        cls = self._choose_by_mode(Grouper, AsyncGrouper)
        self.streamlets.append(cls(self.streamlets[-1], key, **kwargs))
        return self

    def batch(self, batch_size: int) -> Self:
        """
        Take elements from an input stream
        and bundle them up into batches up to a size limit,
        and produce the batches as lists.

        The output batches are all of the specified size,
        except possibly the final batch.
        There is no 'timeout' logic to proceed eagerly with a partial batch.
        For efficiency, this requires the input stream to have a steady supply.
        If that is a concern, having a :meth:`buffer` on the input stream
        prior to :meth:`batch` may help.

        Examples
        --------
        >>> ss = Stream(range(10)).batch(3)
        >>> print(ss.collect())
        [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]
        """
        cls = self._choose_by_mode(Batcher, AsyncBatcher)
        self.streamlets.append(cls(self.streamlets[-1], batch_size))
        return self

    def unbatch(self) -> Self:
        """
        Turn a stream of lists into a stream of individual elements.

        This is sometimes used to correspond with a previous
        :meth:`batch`, but that is by no means a requirement. The only requirement
        is that the input elements are lists.

        ``unbatch`` can be combined with :meth:`map` to implement "expanding" a stream, like this:

        >>> def explode(x):
        ...     if isinstance(x, int) and x > 0:
        ...         return [x] * x
        ...     return []
        >>> ss = Stream([0, 1, 2, 'a', -1, 'b', 3, 4]).map(explode).unbatch()
        >>> print(ss.collect())
        [1, 2, 2, 3, 3, 3, 4, 4, 4, 4]

        In fact, elements of the input stream do not have to be ``list``\\s.
        They can be any `Iterable`_. For example:

        >>> def expand(n):
        ...     for _ in range(n):
        ...         yield n
        >>>
        >>> stream = Stream((1, 2, 4, 3, 0, 5)).map(expand).unbatch().collect()
        >>> print(stream)
        [1, 2, 2, 4, 4, 4, 4, 3, 3, 3, 5, 5, 5, 5, 5]
        """
        cls = self._choose_by_mode(Unbatcher, AsyncUnbatcher)
        self.streamlets.append(cls(self.streamlets[-1]))
        return self

    def accumulate(
        self, func: Callable[[Any, T], Any], initializer: Any = NOTSET, **kwargs
    ) -> Self:
        """
        This method is like "cumulative sum", but the operation is specified by ``func``, hence
        does not need to be "sum". If the last element in the output stream is ``x``
        and the upcoming element in the input stream is ``y``, then the next element in the output
        stream is

        ::

            func(x, y, **kwargs)

        If ``initializer`` is not provided, then the first element is output as is, and "accumulation" begins
        with the second element. Suppose the input stream is ``x0``, ``x1``, ``x2``, ..., then the output
        stream is

        ::

            x0, func(x0, x1, **kwargs), func(func(x0, x1, **kwargs), x2), ...

        If ``initializer`` is provided (any user-provided value, including ``None``), then the first element
        of the output stream is

        ::

            func(initializer, x0, **kwargs)

        hence "accumulation" begins with the first element.

        .. note:: In some other languages or libraries, ``accumulate`` takes a stream or sequence and returns
            a single value as the result. This method, in contrast, returns a value for each element in the stream.
            In fact, the implementation is a simple application of :meth:`map`.

        Examples
        --------
        >>> ss = Stream(range(7))
        >>> print(ss.accumulate(lambda x, y: x + y, 3).collect())
        [3, 4, 6, 9, 13, 18, 24]
        """

        class Accumulator:
            def __init__(self):
                self._func = func
                self._initializer = initializer
                self._kwargs = kwargs

            def __call__(self, x):
                z = self._initializer
                if z is NOTSET:
                    z = x
                else:
                    z = self._func(z, x, **self._kwargs)
                self._initializer = z
                return z

        return self.map(Accumulator())

    def buffer(self, maxsize: int) -> Self:
        """Buffer is used to stabilize and improve the speed of data flow.

        A buffer is useful following any operation that can not guarantee
        (almost) instant availability of output. A buffer allows its
        input to "pile up" when its downstream consumer is slow,
        so that data *is* available when the downstream does come to request
        data. The buffer evens out irregularities in the speeds of upstream
        production and downstream consumption.

        ``maxsize`` is the size of the internal buffer.
        """
        cls = self._choose_by_mode(Buffer, AsyncBuffer)
        self.streamlets.append(cls(self.streamlets[-1], maxsize))
        return self

    def parmap(
        self,
        func: Callable[[T], TT],
        /,
        *,
        num_workers: int | None = None,
        return_x: bool = False,
        return_exceptions: bool = False,
        _async: bool | None = None,
        **kwargs,
    ) -> Self:
        """Parallel, or concurrent, counterpart of :meth:`map`.

        New threads or processes are created to execute ``func``.
        The function is applied on each element of the data stream and produces a new value,
        which forms the output stream.

        Elements in the output stream are in the order of the input elements.
        In other words, the order of data elements is preserved.

        The main difference between :meth:`parmap` and :meth:`map` is that the former
        executes the function concurrently in background threads or processes,
        whereas the latter executes a (simple) function in-line.

        Parameters
        ----------
        func
            A worker function that takes a single input item
            as the first positional argument and produces a result.
            Additional keyword args can be passed in via ``**kwargs``.

            The main point of ``func`` does not have to be its output.
            It could rather be some side effect. For example,
            saving data in a database. In that case, the output may be
            ``None``. Regardless, the output is yielded to be consumed by the next
            operator in the pipeline. A stream of ``None``\\s could be used
            in counting, for example.
        num_workers
            When ``func`` is sync, this is the
            max number of threads created to run ``func``.
            This is also the max number of concurrent calls to ``func``
            that can be ongoing at any time.
            If ``None``, a default value is used.
            Unless there are concrete reasons that you need to restrict the level of concurrency
            or the use of resources, it is recommended to leave `num_workers` at `None`.

            When ``func`` is async, this is the max number of concurrent
            (i.e. ongoing at the same time) calls to ``func``.
        return_x
            If ``True``, output stream will contain tuples ``(x, y)``;
            if ``False``, output stream will contain ``y`` only.
        return_exceptions
            If ``True``, exceptions raised by ``func`` will be
            in the output stream as if they were regular results; if ``False``,
            they will halt the operation and propagate.

            Note that a ``True`` value does not absorb exceptions
            raised by *previous* operators in the pipeline; it is concerned about
            exceptions raised by ``func`` only.
        """
        if (_async is None and inspect.iscoroutinefunction(func)) or (_async is True):
            if isasynciterable(self):
                # This method is called within an async environment.
                # The operator runs in the current thread on the current asyncio event loop.
                cls = AsyncParmapperAsync
            else:
                # Usually this method is called within a sync environment.
                # The operator uses a worker thread to run the async ``func``.
                cls = ParmapperAsync
        else:
            if isasynciterable(self):
                cls = AsyncParmapper
            else:
                cls = Parmapper

        self.streamlets.append(
            cls(
                self.streamlets[-1],
                func,
                num_workers=num_workers,
                return_x=return_x,
                return_exceptions=return_exceptions,
                **kwargs,
            )
        )
        return self


class SyncIter(Iterable):
    def __init__(self, instream: AsyncIterable):
        self._instream = instream

    def _worker(self):
        async def main():
            q = self._q
            to_stop = self._stopped
            try:
                async for x in self._instream:
                    if to_stop.is_set():
                        while True:
                            try:
                                q.get_nowait()
                            except queue.Empty:
                                break
                        return
                    q.put(x)
                q.put(FINISHED)
            except Exception as e:
                q.put(STOPPED)
                q.put(e)

        asyncio.run(main())
        # Since `self._instream` is async iterable,
        # this program is running in an async environment,
        # hence an event loop is running.
        # `asyncio.run_coroutine_threadsafe` might be useful,
        # but I couldn't make it work.

    def _start(self):
        self._q = queue.Queue(2)
        self._stopped = threading.Event()
        self._worker_thread = Thread(target=self._worker)
        self._worker_thread.start()

    def _finalize(self):
        if self._stopped is None:
            return
        self._stopped.set()
        self._worker_thread.join()
        self._stopped = None

    def __iter__(self):
        if isiterable(self._instream):
            yield from self._instream
        else:
            self._start()
            q = self._q
            finished = FINISHED
            stopped = STOPPED
            try:
                while True:
                    x = q.get()
                    if x == finished:
                        break
                    if x == stopped:
                        raise q.get()
                    yield x
            finally:
                self._finalize()


class AsyncIter(AsyncIterable):
    def __init__(self, instream: Iterable):
        self._instream = instream

    async def __aiter__(self):
        if isasynciterable(self._instream):
            async for x in self._instream:
                yield x
        else:
            loop = asyncio.get_running_loop()
            instream = iter(self._instream)
            finished = FINISHED
            while True:
                # ``next(instream)`` could involve some waiting and sleeping,
                # hence doing it in another thread.
                x = await loop.run_in_executor(None, next, instream, finished)
                # `FINISHED` is returned if there's no more elements.
                # See https://stackoverflow.com/a/61774972
                if x == finished:  # `instream_` exhausted
                    break
                yield x


class Mapper(Iterable):
    def __init__(self, instream: Iterable, func: Callable[[T], Any], **kwargs):
        self._instream = instream
        self.func = functools.partial(func, **kwargs) if kwargs else func

    def __iter__(self):
        func = self.func
        for v in self._instream:
            yield func(v)


class AsyncMapper(AsyncIterable):
    def __init__(
        self,
        instream: AsyncIterable,
        func: Callable[[T], Any] | Callable[[T], Awaitable[T]],
        **kwargs,
    ):
        self._instream = instream
        if kwargs:
            func = functools.partial(func, **kwargs)
        self.func = func

    async def __aiter__(self):
        func = self.func
        if iscoroutinefunction(func):
            async for v in self._instream:
                yield await func(v)
        else:
            async for v in self._instream:
                yield func(v)


class Filter(Iterable):
    def __init__(self, instream: Iterable, func: Callable[[T], bool], **kwargs):
        self._instream = instream
        self.func = functools.partial(func, **kwargs) if kwargs else func

    def __iter__(self):
        func = self.func
        for v in self._instream:
            if func(v):
                yield v


class AsyncFilter(AsyncIterable):
    def __init__(
        self,
        instream: AsyncIterable,
        func: Callable[[T], bool] | Callable[[T], Awaitable[bool]],
        **kwargs,
    ):
        self._instream = instream
        self.func = functools.partial(func, **kwargs) if kwargs else func

    async def __aiter__(self):
        func = self.func
        if iscoroutinefunction(func):
            async for v in self._instream:
                if await func(v):
                    yield v
        else:
            async for v in self._instream:
                if func(v):
                    yield v


class Header(Iterable):
    def __init__(self, instream: Iterable, /, n: int):
        """
        Keeps the first ``n`` elements and ignores all the rest.
        """
        assert n > 0
        self._instream = instream
        self.n = n

    def __iter__(self):
        n = 0
        nn = self.n
        for v in self._instream:
            if n >= nn:
                break
            yield v
            n += 1


class AsyncHeader(AsyncIterable):
    def __init__(self, instream: AsyncIterable, /, n: int):
        assert n > 0
        self._instream = instream
        self.n = n

    async def __aiter__(self):
        n = 0
        nn = self.n
        async for v in self._instream:
            if n >= nn:
                break
            yield v
            n += 1


class Tailor(Iterable):
    def __init__(self, instream: Iterable, /, n: int):
        """
        Keeps the last ``n`` elements and ignores all the previous ones.
        If there are less than ``n`` data elements in total, then keep all of them.

        This object needs to walk through the entire input stream before
        starting to yield elements.

        .. note:: ``n`` data elements need to be kept in memory, hence ``n`` should
            not be "too large" for the typical size of the data elements.
        """
        self._instream = instream
        assert n > 0
        self.n = n

    def __iter__(self):
        data = deque(maxlen=self.n)
        for v in self._instream:
            data.append(v)
        yield from data


class AsyncTailor(AsyncIterable):
    def __init__(self, instream: AsyncIterable, /, n: int):
        self._instream = instream
        assert n > 0
        self.n = n

    async def __aiter__(self):
        data = deque(maxlen=self.n)
        async for v in self._instream:
            data.append(v)
        for x in data:
            yield x


class Grouper(Iterable):
    def __init__(self, instream: Iterable, /, key: Callable[[T], Any], **kwargs):
        self._instream = instream
        if kwargs:
            key = functools.partial(key, **kwargs)
        self.key = key

    def __iter__(self):
        yield from itertools.groupby(self._instream, self.key)

        # _z = object()
        # group = None
        # func = self.func
        # for x in self._instream:
        #     z = func(x)
        #     if z == _z:
        #         group.append(x)
        #     else:
        #         if group is not None:
        #             yield _z, group
        #         group = [x]
        #         _z = z
        # if group:
        #     yield _z, group


class AsyncGrouper(AsyncIterable):
    def __init__(
        self,
        instream: AsyncIterable,
        /,
        key: Callable[[T], Any] | Callable[[T], Awaitable[Any]],
        **kwargs,
    ):
        self._instream = instream
        if kwargs:
            key = functools.partial(key, **kwargs)
        self.key = key

    async def __aiter__(self):
        async for v in asyncstdlib.itertools.groupby(self._instream, self.key):
            yield v

        # _z = object()
        # group = None
        # func = self.func
        # async for x in self._instream:
        #     z = func(x)
        #     if z == _z:
        #         group.append(x)
        #     else:
        #         if group is not None:
        #             yield _z, group
        #         group = [x]
        #         _z = z
        # if group:
        #     yield _z, group


class Batcher(Iterable):
    """
    See :meth:`Stream.batch`.
    """

    def __init__(self, instream: Iterable, /, batch_size: int):
        self._instream = instream
        assert batch_size > 0
        self._batch_size = batch_size

    def __iter__(self):
        batch_size = self._batch_size
        batch = []
        for x in self._instream:
            batch.append(x)
            if len(batch) == batch_size:
                yield batch
                batch = []
        if batch:
            yield batch


class AsyncBatcher(AsyncIterable):
    def __init__(self, instream: AsyncIterable, /, batch_size: int):
        self._instream = instream
        assert batch_size > 0
        self._batch_size = batch_size

    async def __aiter__(self):
        batch_size = self._batch_size
        batch = []
        async for x in self._instream:
            batch.append(x)
            if len(batch) == batch_size:
                yield batch
                batch = []
        if batch:
            yield batch


class Unbatcher(Iterable):
    """
    See :meth:`Stream.unbatch`.
    This is comparable to the standard ``itertools.chain.from_iterable``.
    """

    def __init__(self, instream: Iterable, /):
        """
        The incoming stream consists of lists.
        This object "expands" or "flattens" the lists into a stream
        of individual elements. Usually, the output stream
        is "longer" than the input stream.

        This may correspond to a "Batcher" operator upstream,
        but that is by no means a requirement.
        """
        self._instream = instream

    def __iter__(self):
        for x in self._instream:
            yield from x


class AsyncUnbatcher(AsyncIterable):
    def __init__(self, instream: AsyncIterable, /):
        self._instream = instream

    async def __aiter__(self):
        async for x in self._instream:
            # `x` must be iterable or async iterable.
            if isiterable(x):
                for y in x:
                    yield y
            else:
                async for y in x:
                    yield y


class Shuffler(Iterable):
    def __init__(self, instream: Iterable, /, buffer_size: int = 1000):
        assert buffer_size > 0
        self._instream = instream
        self._buffersize = buffer_size

    def __iter__(self):
        buffer = []
        buffersize = self._buffersize
        randrange = random.randrange
        for x in self._instream:
            if len(buffer) < buffersize:
                buffer.append(x)
            else:
                idx = randrange(buffersize)
                y = buffer[idx]
                buffer[idx] = x
                yield y
        if buffer:
            random.shuffle(buffer)
            yield from buffer


class AsyncShuffler(AsyncIterable):
    def __init__(self, instream: AsyncIterable, /, buffer_size: int = 1000):
        assert buffer_size > 0
        self._instream = instream
        self._buffersize = buffer_size

    async def __aiter__(self):
        buffer = []
        buffersize = self._buffersize
        randrange = random.randrange
        async for x in self._instream:
            if len(buffer) < buffersize:
                buffer.append(x)
            else:
                idx = randrange(buffersize)
                y = buffer[idx]
                buffer[idx] = x
                yield y
        if buffer:
            random.shuffle(buffer)
            for x in buffer:
                yield x


class Buffer(Iterable):
    def __init__(self, instream: Iterable, /, maxsize: int):
        self._instream = instream
        assert 1 <= maxsize <= 10_000
        self.maxsize = maxsize

    def _start(self):
        self._stopped = threading.Event()
        self._tasks = SingleLane(self.maxsize)
        self._worker = Thread(target=self._run_worker, name='Buffer-worker-thread')
        self._worker.start()

    def _run_worker(self):
        q = self._tasks
        try:
            for x in self._instream:
                if self._stopped.is_set():
                    break
                q.put(x)  # if `q` is full, will wait here
            q.put(FINISHED)
        except Exception as e:
            q.put(STOPPED)
            q.put(e)
            # raise
            # Do not raise here. Otherwise it would print traceback,
            # while the same would be printed again in ``__iter__``.

    def _finalize(self):
        if self._stopped is None:
            return
        self._stopped.set()
        tasks = self._tasks
        while not tasks.empty():
            _ = tasks.get()
        # `tasks` is now empty. The thread needs to put at most one
        # more element into the queue, which is safe.
        self._worker.join()
        self._stopped = None

    def __iter__(self):
        self._start()
        tasks = self._tasks
        finished = FINISHED
        stopped = STOPPED
        try:
            while True:
                z = tasks.get()
                if z == finished:
                    break
                if z == stopped:
                    raise tasks.get()
                yield z
        finally:
            self._finalize()


class AsyncBuffer(AsyncIterable):
    def __init__(self, instream: AsyncIterable, /, maxsize: int):
        self._instream = instream
        assert 1 <= maxsize <= 10_000
        self.maxsize = maxsize

    def _start(self):
        self._stopped = threading.Event()
        self._tasks = SingleLane(self.maxsize)
        self._worker = Thread(target=self._run_worker, name='AsyncBuffer-worker-thread')
        self._worker.start()

    def _run_worker(self):
        async def main():
            q = self._tasks
            try:
                async for x in self._instream:
                    if self._stopped.is_set():
                        break
                    q.put(x)  # if `q` is full, will wait here
                q.put(FINISHED)
            except Exception as e:
                q.put(STOPPED)
                q.put(e)
                # raise
                # Do not raise here. Otherwise it would print traceback,
                # while the same would be printed again in ``__iter__``.

        asyncio.run(main())

    def _finalize(self):
        if self._stopped is None:
            return
        self._stopped.set()
        tasks = self._tasks
        while not tasks.empty():
            _ = tasks.get()
        # `tasks` is now empty. The thread needs to put at most one
        # more element into the queue, which is safe.
        self._worker.join()
        self._stopped = None

    async def __aiter__(self):
        self._start()
        tasks = self._tasks
        finished = FINISHED
        stopped = STOPPED
        try:
            while True:
                try:
                    z = tasks.get_nowait()
                except queue.Empty:
                    await asyncio.sleep(0.002)
                    # TODO: how to avoid this sleep?
                    continue
                if z == finished:
                    break
                if z == stopped:
                    raise tasks.get()
                yield z
        finally:
            self._finalize()


class ParmapperMixin:
    def _start(self):
        num_workers = self._num_workers
        self._executor_is_shared = False

        if self._executor_type == 'thread':
            self._stopped = threading.Event()
            if num_workers is None and self._executor_initializer is None:
                self._executor = get_shared_thread_pool('_mpservice_streamer_')
                self._executor_is_shared = True
            else:
                self._executor = ThreadPoolExecutor(
                    num_workers,
                    initializer=self._executor_initializer,
                    initargs=self._executor_init_args,
                    thread_name_prefix=self._name + '-thread',
                )
        else:
            self._stopped = Event()
            if num_workers is None and self._executor_initializer is None:
                self._executor = get_shared_process_pool('_mpservice_streamer_')
                self._executor_is_shared = True
            else:
                self._executor = ProcessPoolExecutor(
                    num_workers,
                    initializer=self._executor_initializer,
                    initargs=self._executor_init_args,
                )
                # TODO: what about the process names?

        num_workers = self._executor._max_workers
        self._tasks = SingleLane(num_workers + 1)
        self._worker = Thread(
            target=self._run_worker,
            name=self._name + '-thread-runner',
        )
        self._worker.start()

    def _finalize(self):
        if self._stopped is None:
            return
        self._stopped.set()

        tasks = self._tasks
        worker = self._worker

        while True:
            while not tasks.empty():
                t = tasks.get()
                if isinstance(t, concurrent.futures.Future):
                    t.cancel()

            worker.join(timeout=0.01)
            if not worker.is_alive():
                break

        self._stopped = None
        if not self._executor_is_shared:
            try:
                self._executor.shutdown()
            except OSError:
                pass


class Parmapper(Iterable, ParmapperMixin):
    # Environ is sync; worker func is sync.
    def __init__(
        self,
        instream: Iterable,
        func: Callable[[T], TT],
        *,
        executor: Literal['thread', 'process'],
        num_workers: int | None = None,
        return_x: bool = False,
        return_exceptions: bool = False,
        executor_initializer=None,
        executor_init_args=(),
        parmapper_name='parmapper-sync-sync',
        **kwargs,
    ):
        """
        Parameters
        ----------
        executor
            Either 'thread' or 'process'.

            If ``executor`` is ``'process'``, then ``func`` must be pickle-able,
            for example, it can't be a lambda or a function defined within
            another function. The same caution applies to any parameter passed
            to ``func`` in ``kwargs``.
        kwargs
            Named arguments to ``func``, in addition to the first, positional
            argument, which is an element of ``instream``.
        """
        assert executor in ('thread', 'process')
        if executor_initializer is None:
            assert not executor_init_args
        self._instream = instream
        self._func = func
        self._func_kwargs = kwargs
        self._return_x = return_x
        self._return_exceptions = return_exceptions
        self._executor_type = executor
        self._num_workers = num_workers
        self._executor_initializer = executor_initializer
        self._executor_init_args = executor_init_args
        self._executor = None
        self._executor_is_shared = None
        self._name = parmapper_name
        self._tasks = None
        self._worker = None
        self._stopped = None

    def _run_worker(self):
        func = self._func
        kwargs = self._func_kwargs
        to_stop = self._stopped
        executor = self._executor
        tasks = self._tasks
        try:
            for x in self._instream:
                if to_stop.is_set():
                    break
                t = executor.submit(func, x, loud_exception=False, **kwargs)
                tasks.put((x, t))
                # The size of the queue `tasks` regulates how many
                # concurrent calls to `func` there can be.
            tasks.put(FINISHED)
        except Exception as e:
            tasks.put(STOPPED)
            tasks.put(e)
            # raise
            # Do not raise here. Otherwise it would print traceback,
            # while the same would be printed again in ``__iter__``.

    def __iter__(self):
        self._start()
        finished = FINISHED
        stopped = STOPPED
        try:
            while True:
                z = self._tasks.get()
                if z == finished:
                    break
                if z == stopped:
                    # raise self._worker.exception()
                    raise self._tasks.get()

                x, fut = z

                try:
                    y = fut.result()
                except Exception as e:
                    if self._return_exceptions:
                        # TODO: think about when `e` is a "remote exception".
                        if self._return_x:
                            yield x, e
                        else:
                            yield e
                    else:
                        raise
                else:
                    if self._return_x:
                        yield x, y
                    else:
                        yield y
        finally:
            self._finalize()

        # Regarding clean-up of generators, see
        #   https://stackoverflow.com/a/30862344/6178706
        #   https://docs.python.org/3.6/reference/expressions.html#generator.close


class AsyncParmapper(AsyncIterable, ParmapperMixin):
    # Environ is async; worker func is sync.
    def __init__(
        self,
        instream: AsyncIterable,
        func: Callable[[T], TT],
        *,
        executor: Literal['thread', 'process'],
        num_workers: int | None = None,
        return_x: bool = False,
        return_exceptions: bool = False,
        executor_initializer=None,
        executor_init_args=(),
        parmapper_name='parmapper-async-sync',
        **kwargs,
    ):
        assert executor in ('thread', 'process')
        if executor_initializer is None:
            assert not executor_init_args
        self._instream = instream
        self._func = func
        self._func_kwargs = kwargs
        self._return_x = return_x
        self._return_exceptions = return_exceptions
        self._executor_type = executor
        self._num_workers = num_workers
        self._executor_initializer = executor_initializer
        self._executor_init_args = executor_init_args
        self._executor = None
        self._executor_is_shared = None
        self._name = parmapper_name
        self._tasks = None
        self._worker = None
        self._stopped = None

    def _run_worker(self):
        async def enqueue():
            func = self._func
            kwargs = self._func_kwargs
            to_stop = self._stopped
            executor = self._executor
            tasks = self._tasks
            try:
                async for x in self._instream:
                    if to_stop.is_set():
                        break
                    t = executor.submit(func, x, loud_exception=False, **kwargs)
                    tasks.put((x, t))
                    # The size of the queue `tasks` regulates how many
                    # concurrent calls to `func` there can be.
                tasks.put(FINISHED)
            except Exception as e:
                tasks.put(STOPPED)
                tasks.put(e)
                # raise
                # Do not raise here. Otherwise it would print traceback,
                # while the same would be printed again in ``__iter__``.

        asyncio.run(enqueue())

    async def __aiter__(self):
        self._start()
        finished = FINISHED
        stopped = STOPPED
        try:
            while True:
                try:
                    z = self._tasks.get_nowait()
                except queue.Empty:
                    await asyncio.sleep(0.005)
                    # TODO: how to avoid this sleep?
                    continue

                if z == finished:
                    break
                if z == stopped:
                    # raise self._worker.exception()
                    raise self._tasks.get()

                x, fut = z

                while not fut.done():
                    await asyncio.sleep(0.005)
                    # TODO: how to avoid this sleep?

                try:
                    y = fut.result()
                except Exception as e:
                    if self._return_exceptions:
                        # TODO: think about when `e` is a "remote exception".
                        if self._return_x:
                            yield x, e
                        else:
                            yield e
                    else:
                        raise
                else:
                    if self._return_x:
                        yield x, y
                    else:
                        yield y

        finally:
            self._finalize()


class ParmapperAsync(Iterable):
    # Environ is sync; worker func is async.

    def __init__(
        self,
        instream: Iterable,
        func: Callable[[T], Awaitable[TT]],
        *,
        num_workers: int | None = None,
        return_x: bool = False,
        return_exceptions: bool = False,
        parmapper_name='parmapper-sync-async',
        async_context: dict = None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        async_context
            This dict contains *async* context managers that are arguments to ``func``,
            in addition to the arguments in ``kwargs``. These objects have been initialized
            (``__init__`` called), but have not entered their contexts yet (``__aenter__`` not called).
            The ``__aenter__`` and ``__aexit__`` of each of these objects will be called
            once  at the beginning and end in the worker thread (properly in an async environment).
            These objects will be passed to ``func`` as named arguments. Inside ``func``,
            these objects are used directly without ``async with``. In other words, each of
            these objects enters their context only once, then is used in any number of
            invokations of ``func``.

            One example: ``func`` conducts HTTP requests using the package ``httpx`` and uses
            a ``httpx.AsyncClient`` as a "session" object. You may do something like this::

                async def download_image(url, *, session: httpx.AsyncClient, **kwargs):
                    response = await session.get(url, **kwargs)
                    return response.content

                stream = Stream(urls)
                stream.parmap_async(download_image, async_context={'session': httpx.AsyncClient()}, **kwargs)
                for img in stream:
                    ...
        """
        self._instream = instream
        self._func = func
        self._func_kwargs = kwargs
        self._return_x = return_x
        self._return_exceptions = return_exceptions
        self._num_workers = num_workers or 256
        self._name = parmapper_name
        self._worker_thread = None
        self._stopped = None
        self._async_context = async_context or {}

    def _start(self):
        self._outstream = queue.Queue(self._num_workers)
        self._stopped = threading.Event()

        self._worker_thread = Thread(
            target=self._run_worker,
            name=f'{self._name}-thread',
            # TODO: what if there are multiple such threads owned by multiple ParmapperAsync objects?
            # How to use diff names for them?
        )
        self._worker_thread.start()

    def _finalize(self):
        if self._stopped is None:
            return
        self._stopped.set()
        self._worker_thread.join()
        self._stopped = None

    def _run_worker(self):
        # In the following async functions, some sync operations like I/O on a ``queue.Queue``
        # could involve some waiting. To prevent them from blocking async task executions,
        # run them in a thread executor.

        async def enqueue(tasks):
            loop = asyncio.get_running_loop()
            func = self._func
            kwargs = {**self._func_kwargs, **self._async_context}
            to_stop = self._stopped
            try:
                instream = AsyncIter(self._instream)
                async for x in instream:
                    if to_stop.is_set():
                        break
                    t = loop.create_task(func(x, **kwargs))
                    await tasks.put((x, t))
                    # The size of the queue `tasks` regulates how many
                    # concurrent calls to `func` there can be.
                await tasks.put(FINISHED)
            except Exception as e:
                await tasks.put(STOPPED)
                await tasks.put(e)
                # raise
                # Do not raise here. Otherwise it would print traceback,
                # while the same would be printed again in ``__iter__``.

        async def dequeue(tasks):
            loop = asyncio.get_running_loop()
            thread_pool = get_shared_thread_pool('_mpservice_streamer_')
            outstream = self._outstream
            to_stop = self._stopped
            return_exceptions = self._return_exceptions
            n_submitted = 0
            finished = FINISHED
            stopped = STOPPED
            while True:
                v = await tasks.get()
                if v == finished:
                    # This is placed by `enqueue`, hence
                    # must be the last item in the queue.
                    try:
                        outstream.put_nowait(finished)
                    except queue.Full:
                        await loop.run_in_executor(thread_pool, outstream.put, finished)
                    # No more cleanup is needed.
                    return
                if v == stopped:
                    try:
                        outstream.put_nowait(stopped)
                    except queue.Full:
                        await loop.run_in_executor(thread_pool, outstream.put, stopped)
                    e = await tasks.get()
                    # This is placed by `enqueue`, hence
                    # must be the last item in the queue.
                    try:
                        outstream.put_nowait(e)
                    except queue.Full:
                        await loop.run_in_executor(thread_pool, outstream.put, e)
                    # No more cleanup is needed.
                    return
                if to_stop.is_set():
                    break
                x, t = v
                n_submitted += 1
                try:
                    y = await t
                except Exception as e:
                    try:
                        outstream.put_nowait((x, e))
                    except queue.Full:
                        await loop.run_in_executor(thread_pool, outstream.put, (x, e))
                    if not return_exceptions:
                        to_stop.set()  # signal `enqueue` to stop
                        break
                else:
                    try:
                        outstream.put_nowait((x, y))
                    except queue.Full:
                        await loop.run_in_executor(thread_pool, outstream.put, (x, y))

            # Stop has been requested.
            # Get all Tasks out of the queue;
            # cancel those that are not finished.
            # Do not cancel and wait for cancellation to finish one by one.
            # Instead, send cancel signal into all of them, then wait on them.
            if not to_stop.is_set():
                raise ValueError(
                    f'expecting `stopped.is_set()` to be True but got: {to_stop.is_set()}'
                )
            cancelling = []
            while True:
                v = await tasks.get()
                if v == finished:
                    break
                if v == stopped:
                    await tasks.get()
                    break
                x, t = v
                n_submitted += 1
                t.cancel()
                cancelling.append(t)
            logger.debug(
                f'cancelling {len(cancelling)} of the {n_submitted} tasks submitted'
            )
            for t in cancelling:
                try:
                    await t
                except (asyncio.CancelledError, Exception):  # noqa: S110
                    pass

        async def main():
            async with contextlib.AsyncExitStack() as stack:
                for cm in self._async_context.values():
                    await stack.enter_async_context(cm)
                tasks = asyncio.Queue(self._num_workers - 2)
                t1 = asyncio.create_task(enqueue(tasks))
                t2 = asyncio.create_task(dequeue(tasks))
                await t1
                await t2

        asyncio.run(main())

    def __iter__(self):
        self._start()
        outstream = self._outstream
        finished = FINISHED
        stopped = STOPPED
        try:
            while True:
                z = outstream.get()
                if z == finished:
                    break
                if z == stopped:
                    raise outstream.get()

                x, y = z
                if isinstance(y, Exception):
                    e = y
                    if self._return_exceptions:
                        if self._return_x:
                            yield x, e
                        else:
                            yield e
                    else:
                        raise e
                else:
                    if self._return_x:
                        yield x, y
                    else:
                        yield y
        finally:
            self._finalize()


class AsyncParmapperAsync(AsyncIterable):
    # Environ is async; worker func is async.
    def __init__(
        self,
        instream: AsyncIterable,
        func: Callable[[T], Awaitable[TT]],
        *,
        num_workers: int | None = None,
        return_x: bool = False,
        return_exceptions: bool = False,
        parmapper_name='parmapper-async-async',
        **kwargs,
    ):
        self._instream = instream
        self._func = func
        self._func_kwargs = kwargs
        self._return_x = return_x
        self._return_exceptions = return_exceptions
        self._num_workers = num_workers or 256
        self._name = parmapper_name
        self._tasks = None
        self._worker = None
        self._n_submitted = None

    async def _start(self):
        async def enqueue():
            instream = self._instream
            tasks = self._tasks
            loop = asyncio.get_running_loop()
            func = self._func
            kwargs = self._func_kwargs
            n_submitted = 0
            try:
                try:
                    async for x in instream:
                        t = loop.create_task(func(x, **kwargs))
                        await tasks.put((x, t))
                        n_submitted += 1
                        # The size of the queue `tasks` regulates how many
                        # concurrent calls to `func` there can be.
                except asyncio.CancelledError:
                    self._n_submitted = n_submitted
                    await tasks.put(FINISHED)
                    raise
                else:
                    self._n_submitted = n_submitted
                    await tasks.put(FINISHED)
            except Exception as e:
                self._n_submitted = n_submitted
                await tasks.put(STOPPED)
                await tasks.put(e)

        self._tasks = asyncio.Queue(self._num_workers - 2)
        self._worker = asyncio.create_task(enqueue(), name=self._name)

    async def _finalize(self, done):
        if self._worker is None:
            return
        self._worker.cancel()
        # At this point `self._worker` could have finished.
        finished = FINISHED
        stopped = STOPPED
        if not done:
            tasks = self._tasks
            tt = []
            while True:
                z = await tasks.get()
                if z == finished:
                    break
                if z == stopped:
                    _ = tasks.get()
                    break
                x, t = z
                t.cancel()
                tt.append(t)
            logger.debug(
                f'cancelling {len(tt)} of the {self._n_submitted} tasks submitted'
            )
            for t in tt:
                try:
                    await t
                except (asyncio.CancelledError, Exception):  # noqa: S110
                    pass
        try:
            await self._worker
        except asyncio.CancelledError:
            pass
        self._worker = None

    async def __aiter__(self):
        await self._start()
        tasks = self._tasks
        done = False
        finished = FINISHED
        stopped = STOPPED
        try:
            while True:
                z = await tasks.get()
                if z == finished:
                    done = True
                    break
                if z == stopped:
                    e = await tasks.get()
                    done = True
                    raise e

                x, t = z
                try:
                    y = await t
                except Exception as e:
                    if self._return_exceptions:
                        if self._return_x:
                            yield x, e
                        else:
                            yield e
                    else:
                        raise e
                else:
                    if self._return_x:
                        yield x, y
                    else:
                        yield y
        finally:
            await self._finalize(done)

        # About termination of async generators, see
        #  https://docs.python.org/3.6/reference/expressions.html#asynchronous-generator-functions
        #
        # and
        #  https://peps.python.org/pep-0525/#finalization
        #  https://docs.python.org/3/library/sys.html#sys.set_asyncgen_hooks
        #  https://snarky.ca/unravelling-async-for-loops/
        #  https://github.com/python/cpython/blob/3.11/Lib/asyncio/base_events.py#L539
        #  https://stackoverflow.com/questions/60226557/how-to-forcefully-close-an-async-generator


class TeeX:
    # Tee element
    __slots__ = ('value', 'next', 'n', 'lock')
    # `n`` is count of the times this element has been "consumed".
    # Once `n` is equal to the number of forks in the tee, this
    # element can be discarded. In that situation, this element must
    # be at the head of the queue.

    def __init__(self, x, /):
        self.value = x
        self.next = None
        self.n = 0
        self.lock = threading.Lock()


class Fork:
    def __init__(
        self,
        instream: Iterator,
        n_forks: int,
        buffer: queue.Queue,
        head: SimpleNamespace,
        instream_lock: threading.Lock,
        fork_idx,
    ):
        self.instream = instream
        self.n_forks = n_forks
        self.buffer = buffer
        self.head = head
        self.instream_lock = instream_lock
        self.next: TeeX | None = None
        self._state = 0
        self._fork_idx = fork_idx

    def __iter__(self):
        return self

    def __next__(self):
        if self.next is None:
            if self.head.value is None:
                with self.instream_lock:
                    if self.head.value is None:
                        # Get the very first data element out of `instream`
                        # across all forks.
                        # If this raises `StopIteration`, meaning `instream`
                        # is empty, the exception will be propagated, halting
                        # this fork. All the other forks will also get to this
                        # point and exit the same way.
                        x = next(self.instream)
                        box = TeeX(x)
                        self.buffer.put(box)
                        self.head.value = box
                self.next = self.head.value
                return self.__next__()
            elif self._state == 0:
                # This fork is getting the first element.
                # Some fork (possibly this fork itself in the block above)
                # has obtained the element and assigned it to `self.head.value`.
                self.next = self.head.value
                return self.__next__()
            else:
                raise StopIteration
        else:
            while self.next.next is None:
                # During this loop while waiting on the `instream_lock`,
                # `self.next.next` may become not None thanks to another Fork's
                # actions.

                # Pre-fetch the next element because `self.next` points to
                # the final data element in the buffer.
                locked = self.instream_lock.acquire(timeout=0.1)
                if locked:
                    if self.next.next is None:
                        try:
                            x = next(self.instream)
                        except StopIteration:
                            # `instream` is exhausted.
                            # `self.next.next` remains `None`.
                            # The next call to `__next__` will land
                            # in the first branch and raise `StopIteration`.
                            pass
                        else:
                            box = TeeX(x)
                            self.next.next = box  # IMPORTANT: this line goes before the next to avoid race.
                            self.buffer.put(box)
                    self.instream_lock.release()
                    break

            # Check whether the buffer head should be popped:
            box = self.next
            with box.lock:
                box.n += 1
                if box.n == self.n_forks:
                    # No other fork would be accessing this TeeX "x" at this moment,
                    # and no other fork would be trying to pop the head of the buffer
                    # at this time, hence the it's OK to continue holding the lock.
                    self.buffer.get()

            self.next = box.next
            self._state = 1
            return box.value


def tee(
    instream: Iterable[Elem], n: int = 2, /, *, buffer_size: int = 256
) -> tuple[Stream[Elem], ...]:
    """
    ``tee`` produces multiple (default 2) "copies" of the input data stream,
    to be used in different ways.

    Suppose we have a data stream, on which we want to apply two different lines of operations, conceptually
    like this::

        data = [...]
        stream_1 = Stream(data).map(...).batch(...).parmap(...)...
        stream_2 = stream(data).buffer(...).parmap(...).map(...)...

    There are a few ways we can do this:

    1. Revise the code to apply multiple operations on each data element, that is, "merging"
       the two lines of operations. This will likely make the code more complex.

    2. Apply the first line of operations; then re-walk the data stream to apply the second
       line of operations. When feasible, this is simple and clean. However, this approach could
       be infeasible, for example, when walking the data stream is expensive.
       There are also stituations where it's not possible to get the data stream a second time.

    3. Use the current function::

        data = [...]
        stream_1, stream_2 = teee(data, 2)

       The two streams can be applied the :class:`Stream` operations freely and independently
       (after all, they are proper ``Stream`` objects), except for the final "trigger" operations
       :meth:`~Stream.__iter__`, :meth:`~Stream.collect`, and :meth:`~Stream.drain`::

        stream_1.map(...).batch(...).parmap(...)...
        stream_2.buffer(...).parmap(...).map(...)...

       Upon this setup, we need to start "consuming" the two streams at once, letting them run
       concurrently, and they will also finish around the same time. The concurrent "trigger"
       needs to be in different threads. Suppose ``stream_1`` is run for some side effect, hence
       we can simply ``drain`` it; for ``stream_2``, on the other hand, suppose we need
       its outcoming elements. We may do this::

        with concurrent.futures.ThreadPoolExecutor() as pool:
            t1 = pool.submit(stream_1.drain)
            t2 = pool.submit(stream_2.collect)
            n = t1.result()
            output = t2.result()

    The opposite of ``tee`` is the built-in ``zip``.

    Parameters
    ----------
    buffer_size
        Size of an internal buffer, in terms of the number of elements of ``instream`` it holds.
        Unless each element takes a lot of memory, it is recommended to use a large buffer size,
        such as the default 256.

        Think of the buffer as a "moving window" on ``instream``. The ``n`` "forks" produced by
        this function are iterating over ``instream`` independently subject to one constraint:
        at any moment, the next element to be obtained by each fork is an element in this window.
        Once the first element in the window has been obtained by all ``n`` forks, the window will
        advance by one element. (In detail, the oldest element in the window is dropped; the next
        element is obtained from ``instream`` and appended to the window as the newest element.)

        This moving window gives the "forks" wiggle room in their consumption of the data---they
        do not have to be processing the same element at the same time; their respective concurrent
        processing may be slow at different elements. A larger buffer window reduces the need
        for the forks to wait for their slower peers (if the slowness is on random elements rather
        than on every element).
    """
    assert buffer_size >= 2
    # In practice, use a reasonably large value that is feasible for the application.

    buffer = queue.Queue(buffer_size)

    if not hasattr(instream, '__next__'):
        instream = iter(instream)
    instream_lock = threading.Lock()

    head = SimpleNamespace()
    head.value = None
    # `head` holds the very first element of `instream`.
    # Once assigned, `head` will not change.

    forks = tuple(Fork(instream, n, buffer, head, instream_lock, i) for i in range(n))
    return tuple(Stream(f) for f in forks)


class EagerBatcher(Iterable):
    """
    ``EagerBatcher`` collects items from the incoming stream towards a target batch size and yields the batches.
    For each batch, after getting the first item, it will yield the batch either it has collected enough items
    or has reached ``timeout``. Note, the timer starts upon getting the first item, whereas getting the first item
    for a new batch may take however long.
    """

    def __init__(
        self,
        instream,
        /,
        batch_size: int,
        timeout: float = None,
        endmarker=None,
    ):
        # ``instream`` is a queue (thread or process) with the specicial value ``endmarker``
        # indicating the end of the stream.
        # ``timeout`` can be 0.
        self._instream = instream
        self._batch_size = batch_size
        if timeout is None:
            timeout = 3600 * 24  # effectively unlimited wait
        self._timeout = timeout
        self._endmarker = endmarker

    def __iter__(self):
        q_in = self._instream
        batchsize = self._batch_size
        timeout = self._timeout
        end = self._endmarker

        while True:
            z = q_in.get()  # wait as long as it takes to get one item.
            if end is None:
                if z is None:
                    break
            else:
                if z == end:
                    break

            batch = [z]
            n = 1
            deadline = perf_counter() + timeout
            # Timeout starts after the first item is obtained.

            while n < batchsize:
                t = deadline - perf_counter()
                try:
                    # If `t <= 0`, still get the next item
                    # if it's already available.
                    # In other words, if data elements are already here,
                    # get more towards the target batch-size
                    # even if it's already past the timeout deadline.
                    z = q_in.get(timeout=max(0, t))
                except queue.Empty:
                    break

                if end is None:
                    if z is None:
                        yield batch
                        return
                else:
                    if z == end:
                        yield batch
                        return

                batch.append(z)
                n += 1

            yield batch



"""
The class :class:`RemoteException` is a pickle helper for Exception objects to preserve some traceback info.
This is needed because directly calling ``pickle.dumps`` on an Exception object will lose
its ``__traceback__`` and ``__cause__`` attributes.

This is designed to be used to send an exception object to another process.
It can also be used to pickle-persist an Exception object for some time.

This class preserves some traceback info simply by keeping it as a formatted string during pickling.
Once unpickled, the object obtained, say ``obj``,
is **not** an instance of ``RemoteException``, but rather of the original Exception type.
``obj`` does not have the ``__traceback__`` attribute, which is impossible to reconstruct,
but rather has the ``__cause__`` attribute, which is a custom Exception
object (:class:`RemoteTraceback`) that contains the string-form traceback info, with proper line breaks.
Most (if not all) methods and attributes of ``obj`` behave the same
as the original Exception object, except that
``__traceback__`` is gone but ``__cause__`` is added.

.. note:: ``type(pickle.loads(pickle.dumps(RemoteException(e))))`` is not ``RemoteException``, but rather ``type(e)``.

This design is a compromise.
Since the object out of unpickling is not an instance of ``RemoteException``,
we "lose control of it", in the sense that we can't add methods in ``RemoteException`` to help on the use of that object.
A clear benefit, though, is that we can detect the type of the exception by ``isinstanceof`` or by the type list in
``try ... except <type_list>``, *without even knowing about* ``RemoteException``.

.. note:: ``RemoteException`` does not subclass ``BaseException``, hence you can't *raise* an instance of this class.

.. seealso:: :func:`is_remote_exception`, :func:`get_remote_traceback`.

The session below shows the basic behaviors of a :class:`RemoteException`.

>>> from mpservice.multiprocessing import RemoteException, is_remote_exception, get_remote_traceback
>>> import pickle
>>>
>>>
>>> def foo():
...     raise ValueError(38)
>>>
>>>
>>> def gee():
...     foo()
>>>
>>>
>>> gee()
Traceback (most recent call last):
    File "<stdin>", line 1, in <module>
    File "<stdin>", line 2, in gee
    File "<stdin>", line 2, in foo
ValueError: 38
.
38
>>>
>>> err = None
>>> try:
...     gee()
... except Exception as e:
...     err = e
>>>
>>> err
ValueError(38)
>>> err.__traceback__  # doctest: +ELLIPSIS
<traceback object at 0x7...>
>>> err.__cause__ is None
True
>>>
>>> e_remote = RemoteException(err)
>>> e_remote
RemoteException(ValueError(38))
>>> e_pickled = pickle.dumps(e_remote)
>>> e_unpickled = pickle.loads(e_pickled)
>>>
>>> e_unpickled
ValueError(38)
>>> type(e_unpickled)
<class 'ValueError'>
>>> e_unpickled.__traceback__ is None
True
>>> e_unpickled.__cause__  # doctest: +SKIP
RemoteTraceback('Traceback (most recent call last):
  File "<stdin>", line 2, in <module>
  File "<stdin>", line 2, in gee
  File "<stdin>", line 2, in foo
ValueError: 38
')
>>>
>>> is_remote_exception(e_unpickled)
True
>>> get_remote_traceback(e_unpickled)  # doctest: +SKIP
'Traceback (most recent call last):
  File "<stdin>", line 2, in <module>
  File "<stdin>", line 2, in gee
  File "<stdin>", line 2, in foo
  ValueError: 38
'
>>> print(get_remote_traceback(e_unpickled))  # doctest: +SKIP
Traceback (most recent call last):
    File "<stdin>", line 2, in <module>
    File "<stdin>", line 2, in gee
    File "<stdin>", line 2, in foo
ValueError: 38
>>>
>>>
>>> raise e_unpickled
Traceback (most recent call last):
    File "<stdin>", line 1, in <module>
ValueError: 38
.
38

Examples
--------
Let's use an example to demonstrate the use of :class:`RemoteException`.
First, create a script with the following content:

.. code-block:: python
    :linenos:

    # error.py
    from mpservice.multiprocessing import MP_SPAWN_CTX, RemoteException

    def increment(qin, qout):
        while True:
            x = qin.get()
            if x is None:
                qout.put(None)
                return
            try:
                qout.put((x, x + 1))
            except Exception as e:
                qout.put((x, e))

    def main():
        qin = MP_SPAWN_CTX.Queue()
        qout = MP_SPAWN_CTX.Queue()
        p = MP_SPAWN_CTX.Process(target=increment, args=(qin, qout))
        p.start()
        qin.put(1)
        qin.put(3)
        qin.put('a')
        qin.put(5)
        qin.put(None)
        p.join()
        while True:
            y = qout.get()
            if y is None:
                break
            print(y)

    if __name__ == '__main__':
        main()

Run it::

    $ python error.py
    (1, 2)
    (3, 4)
    ('a', TypeError('can only concatenate str (not "int") to str'))
    (5, 6)

Everything is as expected. Now, instead, we want to stop the program upon errors, so we change
line 30 to

::

        if isinstance(y[1], BaseException):
            raise y[1]
        print(y)

Note the line numbers at the bottom increase a bit. Now ``raise y[1]`` is on line 32, while the ``main()`` call is line 37.
Run it again::

    $ python error.py
    (1, 2)
    (3, 4)
    Traceback (most recent call last):
    File "error.py", line 35, in <module>
        main()
    File "error.py", line 31, in main
        raise y[1]
    TypeError: can only concatenate str (not "int") to str

Looks good?

Not really. The traceback says the exception happened on line 32 with ``raise y[1]``.
That's not very useful. We get no info about where it **actually** happened.

What's the problem?
Well, on line 13, the ``TypeError`` object ``e`` gets put in a multiprocessing queue, pickled;
on line 27, the object gets taken out of the queue, unpickled. In the process,
``pickle.dumps(e)`` strips off the
attribute ``e.__traceback__`` because traceback is not picklable!

One solution is to change line 13 to ``qout.put((x, RemoteException(e)))``.
Now run it again,

::

    $ python error.py
    (1, 2)
    (3, 4)
    mpservice._remote_exception.RemoteTraceback: Traceback (most recent call last):
    File "/home/docker-user/mpservice/tests/experiments/error.py", line 11, in increment
        qout.put((x, x + 1))
    TypeError: can only concatenate str (not "int") to str


    The above exception was the direct cause of the following exception:

    Traceback (most recent call last):
    File "error.py", line 35, in <module>
        main()
    File "error.py", line 31, in main
        raise y[1]
    TypeError: can only concatenate str (not "int") to str

This time, we get the exact line number where the error actually happened in the child process.

If we need to pass an Exception object through multiple processes, we need to remember that
an ``RemoteException`` object pickled and then unpickled gives rise to an object of the original
``Exception`` class, *not* of ``RemoteException``.
For any Exception object, follow this rule:

.. note:: Before putting any ``Exception`` object in a multiprocessing Queue, wrap it by ``RemoteException``.

The script is revised to show this idea:

.. code-block:: python
    :linenos:

    # error.py
    from mpservice.multiprocessing import MP_SPAWN_CTX, RemoteException

    def increment(qin, qout):
        while True:
            x = qin.get()
            if x is None:
                qout.put(None)
                return
            try:
                qout.put((x, x + 1))
            except Exception as e:
                qout.put((x, RemoteException(e)))

    def worker(qin, qout):
        q = MP_SPAWN_CTX.Queue()
        p = MP_SPAWN_CTX.Process(target=increment, args=(qin, q))
        p.start()
        while True:
            y = q.get()
            if y is None:
                qout.put(y)
                break
            # ... do other things ...
            if isinstance(y[1], BaseException):
                qout.put((y[0], RemoteException(y[1])))
            else:
                qout.put(y)
        p.join()

    def main():
        qin = MP_SPAWN_CTX.Queue()
        qout = MP_SPAWN_CTX.Queue()
        p = MP_SPAWN_CTX.Process(target=worker, args=(qin, qout))
        p.start()
        qin.put(1)
        qin.put(3)
        qin.put('a')
        qin.put(5)
        qin.put(None)
        p.join()
        while True:
            y = qout.get()
            if y is None:
                break
            if isinstance(y[1], BaseException):
                raise y[1]
            print(y)

    if __name__ == '__main__':
        main()

Run it::

    $ python error.py
    (1, 2)
    (3, 4)
    mpservice._remote_exception.RemoteTraceback: Traceback (most recent call last):
    File "/home/docker-user/mpservice/tests/experiments/error.py", line 11, in increment
        qout.put((x, x + 1))
    TypeError: can only concatenate str (not "int") to str


    The above exception was the direct cause of the following exception:

    Traceback (most recent call last):
    File "error.py", line 51, in <module>
        main()
    File "error.py", line 47, in main
        raise y[1]
    TypeError: can only concatenate str (not "int") to str
"""

from __future__ import annotations

import traceback
from types import TracebackType
from typing import Optional

__call__ = [
    'RemoteException',
    'get_remote_traceback',
    'is_remote_exception',
]


# This class should be in the module `mpserver`.
# It is here because the class `RemoteException` needs to handle it.
# User should import it from `mpserver`.
class EnsembleError(RuntimeError):
    def __init__(self, results: dict):
        nerr = sum(
            1 if isinstance(v, (BaseException, RemoteException)) else 0
            for v in results['y']
        )
        errmsg = None
        for v in results['y']:
            if isinstance(v, (BaseException, RemoteException)):
                errmsg = repr(v)
                break
        msg = f"{results['n']}/{len(results['y'])} ensemble members finished, with {nerr} error{'s' if nerr > 1 else ''}; first error: {errmsg}"
        super().__init__(msg, results)
        # self.args[1] is the results

    def __repr__(self):
        return self.args[0]

    def __str__(self):
        return self.args[0]

    def __reduce__(self):
        return type(self), (self.args[1],)


def is_remote_exception(e) -> bool:
    """Return ``True`` if the exception ``e`` was created by :class:`RemoteException`, and ``False`` otherwise."""
    return isinstance(e, BaseException) and isinstance(e.__cause__, RemoteTraceback)


def get_remote_traceback(e) -> str:
    """
    ``e`` must have checked ``True`` by :func:`is_remote_exception`.
    Suppose an Exception object is wrapped by :class:`RemoteException` and sent to another process
    through a queue.
    Taken out of the queue in the other process, the object will check ``True``
    by :func:`is_remote_exception`. Then this function applied on the object
    will return the traceback info as a string.
    """
    return e.__cause__.tb


class RemoteTraceback(Exception):
    """
    This class is used by :class:`RemoteException`.
    End-user does not use this class directly.
    """

    def __init__(self, tb: str):
        self.tb = tb

    def __str__(self):
        return self.tb


def _rebuild_exception(exc: BaseException, tb: str):
    exc.__cause__ = RemoteTraceback(tb)

    return exc


class RemoteException:
    # This takes the idea of `concurrent.futures.process._ExceptionWithTraceback`
    # with slightly tweaked traceback printout.
    # `pebble.common` uses the same idea.

    # check out
    #   https://github.com/ionelmc/python-tblib
    #   https://stackoverflow.com/questions/6126007/python-getting-a-traceback-from-a-multiprocessing-process
    # about pickling Exceptions with tracebacks.
    #
    # See also: boltons.tbutils
    # See also: package `eliot`: https://github.com/itamarst/eliot/blob/master/eliot/_traceback.py
    #
    # Also check out `sys.excepthook`.

    def __init__(self, exc: BaseException, tb: Optional[TracebackType | str] = None):
        """
        Parameters
        ----------
        exc
            A BaseException object.
        tb
            Optional traceback info, if not already carried by ``exc``.
        """
        if isinstance(tb, str):
            pass
        elif isinstance(tb, TracebackType):
            tb = ''.join(traceback.format_exception(type(exc), exc, tb))
        else:
            if tb is not None:
                raise ValueError(f'expecting no traceback but got: {tb}')

            if exc.__traceback__ is not None:
                # This is the most common use case---in an exception handler block:
                #
                #   try:
                #       ...
                #   except Exception as e:
                #       ...
                #       ee = RemoteException(e)
                #       ...
                #
                # This includes the case where `e` has come from another process via `RemoteException`
                # (hence `is_remote_exception(e)` is True) and is raised again, and because
                # we intend to pickle it again (e.g. paassing it to another process via a queue),
                # hence we put it in `RemoteException`.
                tb = ''.join(
                    traceback.format_exception(type(exc), exc, exc.__traceback__)
                )
            else:
                # This use case is not in an "except" block, rather somehow there's an
                # exception object and we need to pickle it, so we put it in a
                # `RemoteException`.
                if is_remote_exception(exc):
                    tb = get_remote_traceback(exc)
                    # `exc.__cause__` will become `None` after pickle/unpickle.
                else:
                    raise ValueError(f'{repr(exc)} does not contain traceback info')
                    # In this case, don't use RemoteException. Pickle the exc object directly.

        if isinstance(exc, EnsembleError):
            # When an EnsembleError is just raised, all the exception objects in it
            # are RemoteException objects, so this block is no op.
            # But once an EnsembleError object has gone through pickling/unpickling,
            # the RemoteException objects contained in it will change back to
            # BaseException objects. If this object is put in RemoteException again
            # (prior to its being placed on a queue), this block will be useful.
            z = exc.args[1]['y']
            for i in range(len(z)):
                if isinstance(z[i], BaseException):
                    z[i] = self.__class__(z[i])

        self.exc = exc
        """
        This is still the original Exception object with traceback and everything.
        When you get a RemoteException object, it must have not gone through pickling
        (because a RemoteException object would not survive pickling!), hence you can
        use its ``exc`` attribute directly.
        """
        self.tb = tb

    def __repr__(self):
        return f'{self.__class__.__name__}({self.exc.__repr__()})'

    def __str__(self):
        return f"{self.__class__.__name__}('{self.exc.__str__()}')"

    def __reduce__(self):
        return _rebuild_exception, (self.exc, self.tb)



from deprecation import deprecated


@deprecated(deprecated_in='0.14.1', removed_in='0.15.0')
def full_class_name(cls):
    if not isinstance(cls, type):
        cls = cls.__class__
    mod = cls.__module__
    if mod is None or mod == 'builtins':
        return cls.__name__
    return mod + '.' + cls.__name__
